# Copyright (C) 2006-2007 Aren Olson
#                    2011 Brian Parma
#                    2020 Rok Mandeljc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import functools
from gettext import gettext as _
import http.client
import logging
import pickle
import os
import time

from gi.repository import Gtk
from gi.repository import GObject

from xl import collection, event, trax, common, providers, settings, xdg
from xlgui.panel.collection import CollectionPanel
from xlgui.widgets import dialogs, menu, menuitems
from xlgui import main

from .client import DAAPClient
from . import daapclientprefs

logger = logging.getLogger(__name__)

_smi = menu.simple_menu_item
_sep = menu.simple_separator


# Check for python-zeroconf
try:
    import zeroconf

    ZEROCONF = True
    ZEROCONF_VERSION = [int(v) for v in zeroconf.__version__.split('.')[:2]]

    # ServiceInfo.parsed_addresses() and IPVersion enum were introduced
    # in v.0.24
    ZEROCONF_LEGACY = ZEROCONF_VERSION < [0, 24]
    if ZEROCONF_LEGACY:
        import socket  # for inet_ntoa
except ImportError:
    ZEROCONF = False


# detect authentication support in python-daap
try:
    tmp = DAAPClient()
    tmp.connect("spam", "eggs", "sausage")  # dummy login
    del tmp
except TypeError:
    AUTH = False
except Exception:
    AUTH = True


class DaapZeroconfInterface(GObject.Object):
    __gsignals__ = {
        'connect': (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_PYOBJECT,))
    }

    def new_share_menu_item(self, menu_name, service_name, address, port):
        """
        This function is called to add a server to the connect menu.
        """

        if not self.menu:
            return

        menu_item = _smi(
            menu_name,
            ['sep'],
            menu_name,
            callback=lambda *_x: self.clicked(service_name, address, port),
        )
        self.menu.add_item(menu_item)

    def clear_share_menu_items(self):
        """
        This function is used to clear all the menu items out of a menu.
        """

        if not self.menu:
            return

        items_to_remove = [
            item
            for item in self.menu._items
            if item.name not in ('manual', 'history', 'sep')
        ]
        for item in items_to_remove:
            self.menu.remove_item(item)

    def rebuild_share_menu_items(self):
        """
        This function fills the menu with known servers.
        """
        self.clear_share_menu_items()

        show_ipv6 = settings.get_option('plugin/daapclient/ipv6', False)
        items = []

        for key, info in self.services.items():
            # Strip the service type from fully-qualified service name
            service_name = info.name
            if service_name.endswith(info.type):
                service_name = service_name[: -(len(info.type) + 1)]

            if ZEROCONF_LEGACY:
                # Legacy mode: returns only a single IPv4 address
                addresses = [socket.inet_ntoa(info.address)]
            else:
                # Retrieve IP address(es)
                if show_ipv6:
                    # Both IPv4 and IPv6
                    addresses = info.parsed_addresses(zeroconf.IPVersion.All)
                else:
                    # IPv4 only
                    addresses = info.parsed_addresses(zeroconf.IPVersion.V4Only)

            # Generate one menu entry for each available address.
            # NOTE: in its current implementation (v.0.25.1), zeroconf
            # appears to always return at most one IPv4 and one IPv6
            # address, even if the service advertises multiple addresses.
            # This appears to be tied to record caching, which keeps
            # track of only the last parsed IPv4 and IPv6 address.
            for address in addresses:
                # gstreamer can't handle link-local ipv6
                if address.startswith('fe80:'):
                    continue

                menu_name = '{0} ({1})'.format(service_name, address)
                items.append((menu_name, service_name, address, info.port))

        # Create menu items
        for item in items:
            self.new_share_menu_item(*item)

    def clicked(self, service_name, address, port):
        """
        This function is called in response to a menu_item click.
        Fire away.
        """
        GObject.idle_add(self.emit, "connect", (service_name, address, port))

    def on_service_state_change(self, service_type, name, state_change, **kwargs):
        # The zeroconf module explicitly passes callback arguments via
        # keywords, and the 'zeroconf' keyword argument clashes with the
        # module name. Hence the ugly work-around via **kwargs...
        zc = kwargs['zeroconf']

        logger.info("DAAP share '{0}': state changed to {1}".format(name, state_change))

        # zeroconf.ServiceStateChange.Updated was introduced in v.0.23
        add_update_states = [zeroconf.ServiceStateChange.Added]
        if hasattr(zeroconf.ServiceStateChange, 'Updated'):
            add_update_states.append(zeroconf.ServiceStateChange.Updated)

        if state_change in add_update_states:
            info = zc.get_service_info(service_type, name)
            if not info:
                return

            self.services[name] = info
        elif state_change is zeroconf.ServiceStateChange.Removed:
            del self.services[name]

        self.rebuild_share_menu_items()

    def __init__(self, _exaile, _menu):
        """
        Sets up the zeroconf listener.
        """
        GObject.Object.__init__(self)
        self.services = {}
        self.menu = _menu

        if ZEROCONF_LEGACY:
            logger.info("Using zeroconf legacy API")
            zc = zeroconf.Zeroconf()
        else:
            logger.info("Using zeroconf new API")
            zc = zeroconf.Zeroconf(ip_version=zeroconf.IPVersion.All)

        self.browser = zeroconf.ServiceBrowser(
            zc, '_daap._tcp.local.', handlers=[self.on_service_state_change]
        )


class DaapHistory(common.LimitedCache):
    def __init__(self, limit=5, location=None, menu=None, callback=None):
        common.LimitedCache.__init__(self, limit)

        if location is None:
            location = os.path.join(xdg.get_cache_dir(), 'daaphistory.dat')
        self.location = location
        self.menu = menu
        self.callback = callback

        self.load()

    def __setitem__(self, item, value):
        common.LimitedCache.__setitem__(self, item, value)

        # add new menu item
        if self.menu is not None and self.callback is not None:
            menu_item = _smi(
                'hist' + item,
                ['sep'],
                item,
                callback=lambda *_x: self.callback(None, value + (None,)),
            )
            self.menu.add_item(menu_item)

    def load(self):
        try:
            with open(self.location, 'rb') as f:
                try:
                    d = pickle.load(f)
                    self.update(d)
                except (IOError, EOFError):
                    # no file
                    pass
        except IOError:
            # file not present
            pass

    def save(self):
        with open(self.location, 'wb') as f:
            pickle.dump(self.cache, f, common.PICKLE_PROTOCOL)


class DaapManager:
    """
        DaapManager is a class that manages DaapConnections, both manual
    and auto-discovered.
    """

    def __init__(self, exaile, _menu, autodiscover):
        """
        Init!  Create manual menu item, and connect to interface signal.
        """
        self.exaile = exaile
        self.autodiscover = autodiscover
        self.panels = {}

        hmenu = menu.Menu(None)

        def hmfactory(_menu, _parent, _context):
            item = Gtk.MenuItem.new_with_mnemonic(_('History'))
            item.set_submenu(hmenu)
            sens = settings.get_option('plugin/daapclient/history', True)
            item.set_sensitive(sens)
            return item

        _menu.add_item(
            _smi('manual', [], _('Manually...'), callback=self.manual_connect)
        )
        _menu.add_item(menu.MenuItem('history', hmfactory, ['manual']))
        _menu.add_item(_sep('sep', ['history']))

        if autodiscover is not None:
            autodiscover.connect("connect", self.connect_share)

        self.history = DaapHistory(5, menu=hmenu, callback=self.connect_share)

    def connect_share(self, obj, args):
        """
        This function is called when a user wants to connec to
        a DAAP share.  It creates a new panel for the share, and
        requests a track list.
        `args` is a tuple of (name, address, port, service)
        """
        name, address, port = args  # unpack tuple
        user_agent = self.exaile.get_user_agent_string(__name__)
        conn = DaapConnection(name, address, port, user_agent)

        conn.connect()
        library = DaapLibrary(conn)
        panel = NetworkPanel(self.exaile.gui.main.window, library, self)
        #    cst = CollectionScanThread(None, panel.net_collection, panel)
        #    cst.start()
        panel.refresh()  # threaded
        providers.register('main-panel', panel)
        self.panels[name] = panel

        # history
        if settings.get_option('plugin/daapclient/history', True):
            self.history[name] = (name, address, port)
            self.history.save()

    def disconnect_share(self, name):
        """
        This function is called to disconnect a previously connected
        share.  It calls the DAAP disconnect, and removes the panel.
        """

        panel = self.panels[name]
        #    panel.library.daap_share.disconnect()
        panel.daap_share.disconnect()
        #    panel.net_collection.remove_library(panel.library)
        providers.unregister('main-panel', panel)
        del self.panels[name]

    def manual_connect(self, *_args):
        """
        This function is called when the user selects the manual
        connection option from the menu.  It requests a host/ip to
        connect to.
        """
        dialog = dialogs.TextEntryDialog(
            _("Enter IP address and port for share"), _("Enter IP address and port.")
        )
        resp = dialog.run()

        if resp == Gtk.ResponseType.OK:
            loc = dialog.get_value().strip()
            host = loc

            # the port will be anything after the last :
            p = host.rfind(":")

            # ipv6 literals should have a closing brace before the port
            b = host.rfind("]")

            if p > b:
                try:
                    port = int(host[p + 1 :])
                    host = host[:p]
                except ValueError:
                    logger.error('non-numeric port specified')
                    return
            else:
                port = 3689  # if no port specified, use default DAAP port

            # if it's an ipv6 host with brackets, strip them
            if host and host[0] == '[' and host[-1] == ']':
                host = host[1:-1]
            self.connect_share(None, (loc, host, port, None))

    def refresh_share(self, name):
        panel = self.panels[name]
        rev = panel.daap_share.session.revision

        # check for changes
        panel.daap_share.session.update()
        logger.debug(
            'DAAP Server %s returned revision %d ( old: %d ) after update request'
            % (name, panel.daap_share.session.revision, rev)
        )

        # if changes, refresh
        if rev != panel.daap_share.session.revision:
            logger.info(
                'DAAP Server %s changed, refreshing... (revision %d)'
                % (name, panel.daap_share.session.revision)
            )
            panel.refresh()

    def close(self, remove=False):
        """
        This function disconnects active DaapConnections, and optionally
        removes the panels from the UI.
        """
        # disconnect active shares
        for panel in self.panels.values():
            panel.daap_share.disconnect()

            # there's no point in doing this if we're just shutting down, only on
            # disable
            if remove:
                providers.unregister('main-panel', panel)


class DaapConnection:
    """
    A connection to a DAAP share.
    """

    def __init__(self, name, server, port, user_agent):
        # if it's an ipv6 address
        if ':' in server and server[0] != '[':
            server = '[' + server + ']'

        self.all = []
        self.session = None
        self.connected = False
        self.tracks = None
        self.server = server
        self.port = port
        self.name = name
        self.auth = False
        self.password = None
        self.user_agent = user_agent

    def connect(self, password=None):
        """
        Connect, login, and retrieve the track list.
        """
        try:
            client = DAAPClient()
            if AUTH and password:
                client.connect(self.server, self.port, password, self.user_agent)
            else:
                client.connect(self.server, self.port, None, self.user_agent)
            self.session = client.login()
            self.connected = True
        #        except DAAPError:
        except Exception:
            logger.exception(
                'failed to connect to ({0},{1})'.format(self.server, self.port)
            )

            self.auth = True
            self.connected = False
            raise

    def disconnect(self):
        """
        Disconnect, clean up.
        """
        try:
            self.session.logout()
        except Exception:
            pass
        self.session = None
        self.tracks = None
        self.database = None
        self.all = []
        self.connected = False

    def reload(self):
        """
        Reload the tracks from the server
        """
        self.tracks = None
        self.database = None
        self.all = []
        self.get_database()

        t = time.time()
        self.convert_list()
        logger.debug('{0} tracks loaded in {1}s'.format(len(self.all), time.time() - t))

    def get_database(self):
        """
        Get a DAAP database and its track list.
        """
        if self.session:
            self.database = self.session.library()
            self.get_tracks(1)

    def get_tracks(self, reset=False):
        """
        Get the track list from a DAAP database
        """
        if reset or self.tracks is None:
            if self.database is None:
                self.database = self.session.library()
            self.tracks = self.database.tracks()

        return self.tracks

    def convert_list(self):
        """
        Converts the DAAP track database into Exaile Tracks.
        """
        # Convert DAAPTrack's attributes to Tracks.
        eqiv = {
            'title': 'minm',
            'artist': 'asar',
            'album': 'asal',
            'tracknumber': 'astn',
            'date': 'asyr',
            'discnumber': 'asdn',
            'albumartist': 'asaa',
        }
        #            'genre':'asgn','enc':'asfm','bitrate':'asbr'}

        for tr in self.tracks:
            if tr is not None:
                # http://<server>:<port>/databases/<dbid>/items/<id>.<type>?session-id=<sessionid>

                uri = "http://%s:%s/databases/%s/items/%s.%s?session-id=%s" % (
                    self.server,
                    self.port,
                    self.database.id,
                    tr.id,
                    tr.type,
                    self.session.sessionid,
                )

                # Don't scan tracks because gio is slow!
                temp = trax.Track(uri, scan=False)

                for field in eqiv.keys():
                    try:
                        tag = '%s' % tr.atom.getAtom(eqiv[field])
                        if tag != 'None':
                            temp.set_tag_raw(field, [tag], notify_changed=False)

                    except Exception:
                        if field == 'tracknumber':
                            temp.set_tag_raw('tracknumber', [0], notify_changed=False)

                # TODO: convert year (asyr) here as well, what's the formula?
                try:
                    temp.set_tag_raw(
                        "__length",
                        tr.atom.getAtom('astm') // 1000,
                        notify_changed=False,
                    )
                except Exception:
                    temp.set_tag_raw("__length", 0, notify_changed=False)

                self.all.append(temp)

    @common.threaded
    def get_track(self, track_id, filename):
        """
        Save the track with track_id to filename
        """
        for t in self.tracks:
            if t.id == track_id:
                try:
                    t.save(filename)
                except http.client.CannotSendRequest:
                    Gtk.MessageDialog(
                        buttons=Gtk.ButtonsType.OK,
                        message_type=Gtk.MessageType.INFO,
                        modal=True,
                        text=_("""This server does not support multiple connections.
You must stop playback before downloading songs."""),
                        transient_for=main.mainwindow().window,
                    )
                    return


class DaapLibrary(collection.Library):
    """
    Library subclass for better management of collection??
    Or something to do with devices or somesuch. Ask Aren.
    """

    def __init__(self, daap_share, col=None):
        #        location = "http://%s:%s/databasese/%s/items/" % (daap_share.server, daap_share.port, daap_share.database.id)
        # Libraries need locations...
        location = "http://%s:%s/" % (daap_share.server, daap_share.port)
        collection.Library.__init__(self, location)
        self.daap_share = daap_share
        # self.collection = col

    def rescan(self, notify_interval=None, force_update=False):
        """
        Called when a library needs to refresh its track list.

        The force_update parameter is not applicable and is ignored.
        """
        if self.collection is None:
            return True

        if self.scanning:
            return
        t = time.time()
        logger.info('Scanning library: %s' % self.daap_share.name)
        self.scanning = True

        # DAAP gives us all the tracks in one dump
        self.daap_share.reload()
        if self.daap_share.all:
            count = len(self.daap_share.all)
        else:
            count = 0

        if count > 0:
            logger.info(
                'Adding %d tracks from %s. (%f s)'
                % (count, self.daap_share.name, time.time() - t)
            )
            self.collection.add_tracks(self.daap_share.all)

        if notify_interval is not None:
            event.log_event('tracks_scanned', self, count)

        # track removal?
        self.scanning = False
        # return True

    # Needed to be overridden for who knows why (exceptions)
    def _count_files(self):
        count = 0
        if self.daap_share:
            count = len(self.daap_share.all)

        return count


class NetworkPanel(CollectionPanel):
    """
    A panel that displays a collection of tracks from the DAAP share.
    """

    def __init__(self, parent, library, mgr):
        """
        Expects a parent Gtk.Window, and a daap connection.
        """

        self.name = str(library.daap_share.name)
        self.daap_share = library.daap_share
        self.net_collection = collection.Collection(self.name)
        self.net_collection.add_library(library)
        CollectionPanel.__init__(
            self,
            parent,
            self.net_collection,
            self.name,
            _show_collection_empty_message=False,
            label=self.name,
        )

        self.all = []

        self.connect_id = None

        self.menu = menu.Menu(self)

        def get_tracks_func(*_args):
            return self.tree.get_selected_tracks()

        self.menu.add_item(menuitems.AppendMenuItem('append', [], get_tracks_func))
        self.menu.add_item(
            menuitems.EnqueueMenuItem('enqueue', ['append'], get_tracks_func)
        )
        self.menu.add_item(
            menuitems.PropertiesMenuItem('props', ['enqueue'], get_tracks_func)
        )
        self.menu.add_item(_sep('sep', ['props']))
        self.menu.add_item(
            _smi(
                'refresh',
                ['sep'],
                _('Refresh Server List'),
                callback=lambda *x: mgr.refresh_share(self.name),
            )
        )
        self.menu.add_item(
            _smi(
                'disconnect',
                ['refresh'],
                _('Disconnect from Server'),
                callback=lambda *x: mgr.disconnect_share(self.name),
            )
        )

    @common.threaded
    def refresh(self):
        """
        This is called to refresh the track list.
        """
        # Since we don't use a ProgressManager/Thingy, we have to call these w/out
        #  a ScanThread
        self.net_collection.rescan_libraries()
        GObject.idle_add(self._refresh_tags_in_tree)

    def save_selected(self, widget=None, event=None):
        """
        Save the selected tracks to disk.
        """
        items = self.get_selected_items()
        dialog = Gtk.FileChooserDialog(
            _("Select a Location for Saving"),
            main.mainwindow().window,
            Gtk.FileChooserAction.SELECT_FOLDER,
            (
                Gtk.STOCK_OPEN,
                Gtk.ResponseType.OK,
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
            ),
        )
        dialog.set_current_folder(xdg.get_last_dir())
        dialog.set_select_multiple(False)
        result = dialog.run()
        dialog.hide()

        if result == Gtk.ResponseType.OK:
            folder = dialog.get_current_folder()
            self.save_items(items, folder)

    @common.threaded
    def save_items(self, items, folder):
        for i in items:
            tnum = i.get_track()
            if tnum < 10:
                tnum = "0%s" % tnum
            else:
                tnum = str(tnum)
            filename = "%s%s%s - %s.%s" % (folder, os.sep, tnum, i.get_title(), i.type)
            i.connection.get_track(i.daapid, filename)


#                print "DAAP: saving track %s to %s."%(i.daapid, filename)


class DaapClientPlugin:
    __exaile = None
    __manager = None

    def enable(self, exaile):
        """
        Plugin Enabled.
        """
        self.__exaile = exaile

    def on_gui_loaded(self):
        event.add_callback(self.__on_settings_changed, 'plugin_daapclient_option_set')

        menu_ = menu.Menu(None)

        providers.register(
            'menubar-tools-menu', _sep('plugin-sep', ['slow-scan-collection'])
        )

        item = _smi('daap', ['plugin-sep'], _('Connect to DAAP...'), submenu=menu_)
        providers.register('menubar-tools-menu', item)

        autodiscover = None
        if ZEROCONF:
            try:
                autodiscover = DaapZeroconfInterface(self.__exaile, menu_)
            except RuntimeError:
                logger.warning('zeroconf interface could not be initialized')
        else:
            logger.warning(
                'python-zeroconf is not available; disabling DAAP share auto-discovery!'
            )

        self.__manager = DaapManager(self.__exaile, menu_, autodiscover)

    def teardown(self, exaile):
        """
        Exaile Shutdown.
        """
        # disconnect from active shares
        if self.__manager is not None:
            self.__manager.close()
            self.__manager = None

    def disable(self, exaile):
        """
        Plugin Disabled.
        """
        self.teardown(exaile)

        for item in providers.get('menubar-tools-menu'):
            if item.name == 'daap':
                providers.unregister('menubar-tools-menu', item)
                break

    def get_preferences_pane(self):
        return daapclientprefs

    def __on_settings_changed(self, event, setting, option):
        if option == 'plugin/daapclient/ipv6' and self.__manager is not None:
            self.__manager.autodiscover.rebuild_share_menu_items()


plugin_class = DaapClientPlugin

# vi: et ts=4 sts=4 sw=4



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import GObject
from gi.repository import Gdk
from gi.repository import Gtk

from xl import event, settings
from xl.common import clamp
from xl.nls import gettext as _

from xlgui import icons


class RatingWidget(Gtk.EventBox):
    """
    A rating widget which displays a row of
    images and allows for selecting the rating
    """

    __gproperties__ = {
        'rating': (
            GObject.TYPE_INT,
            'rating',
            'The selected rating',
            0,  # Minimum
            65535,  # Maximum
            0,  # Default
            GObject.ParamFlags.READWRITE,
        )
    }
    __gsignals__ = {
        'rating-changed': (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_INT,))
    }

    def __init__(self, rating=0, player=None):
        """
        :param rating: the optional initial rating
        :type rating: int
        :param player: If not None, this rating widget will automatically
                       update to reflect the rating of the current song
        :type player: xl.player.ExailePlayer
        """
        Gtk.EventBox.__init__(self)
        self._player = player

        self.set_visible_window(False)
        self.set_above_child(True)
        self.add_events(Gdk.EventMask.POINTER_MOTION_MASK)
        self.set_can_focus(True)

        self._image = Gtk.Image()
        self.add(self._image)

        self._rating = -1
        self.props.rating = rating

        if self._player is not None:
            event.add_ui_callback(
                self.on_rating_update, 'playback_track_start', self._player
            )
            event.add_ui_callback(
                self.on_rating_update, 'playback_track_end', self._player
            )
            event.add_ui_callback(self.on_rating_update, 'rating_changed')

            self.on_rating_update('rating_changed', None, None)

    def destroy(self):
        """
        Cleanups
        """
        if self._player is not None:
            event.remove_callback(
                self.on_rating_update, 'playback_track_start', self._player
            )
            event.remove_callback(
                self.on_rating_update, 'playback_track_end', self._player
            )
            event.remove_callback(self.on_rating_update, 'rating_changed')

    def do_get_property(self, property):
        """
        Getter for custom properties
        """
        if property.name == 'rating':
            return self._rating
        else:
            raise AttributeError('unknown property %s' % property.name)

    def do_set_property(self, property, value):
        """
        Setter for custom properties
        """
        if property.name == 'rating':
            if value == self._rating:
                value = 0
            else:
                value = clamp(value, 0, settings.get_option('rating/maximum', 5))

            self._rating = value
            self._image.set_from_pixbuf(icons.MANAGER.pixbuf_from_rating(value).pixbuf)

            self.emit('rating-changed', value)
        else:
            raise AttributeError('unknown property %s' % property.name)

    def do_expose_event(self, event):
        """
        Takes care of painting the focus indicator
        """
        if self.is_focus():
            self.style.paint_focus(
                window=self.window,
                state_type=self.get_state(),
                area=event.area,
                widget=self,
                detail='button',  # Borrow style from GtkButton
                x=event.area.x,
                y=event.area.y,
                width=event.area.width,
                height=event.area.height,
            )

        Gtk.EventBox.do_expose_event(self, event)

    def do_motion_notify_event(self, event):
        """
        Temporarily updates the displayed rating
        """
        if self.get_state_flags() & Gtk.StateType.INSENSITIVE:
            return

        allocation = self.get_allocation()
        maximum = settings.get_option('rating/maximum', 5)
        pixbuf_width = self._image.get_pixbuf().get_width()
        # Activate pixbuf if half of it has been passed
        threshold = (pixbuf_width / maximum) / 2
        position = (event.x + threshold) / allocation.width
        rating = int(position * maximum)

        self._image.set_from_pixbuf(icons.MANAGER.pixbuf_from_rating(rating).pixbuf)

    def do_leave_notify_event(self, event):
        """
        Restores the original rating
        """
        if self.get_state_flags() & Gtk.StateType.INSENSITIVE:
            return

        self._image.set_from_pixbuf(
            icons.MANAGER.pixbuf_from_rating(self._rating).pixbuf
        )

    def do_button_release_event(self, event):
        """
        Applies the selected rating
        """
        if self.get_state_flags() & Gtk.StateType.INSENSITIVE:
            return

        allocation = self.get_allocation()
        maximum = settings.get_option('rating/maximum', 5)
        pixbuf_width = self._image.get_pixbuf().get_width()
        # Activate pixbuf if half of it has been passed
        threshold = (pixbuf_width / maximum) / 2
        position = (event.x + threshold) / allocation.width
        self.props.rating = int(position * maximum)

    def do_key_press_event(self, event):
        """
        Changes the rating on keyboard interaction
        * Alt+Up/Right: increases the rating
        * Alt+Down/Left: decreases the rating
        """
        if self.get_state_flags() & Gtk.StateType.INSENSITIVE:
            return

        if not event.get_state() & Gdk.ModifierType.MOD1_MASK:
            return

        if event.keyval in (Gdk.KEY_Up, Gdk.KEY_Right):
            rating = self.props.rating + 1
        elif event.keyval in (Gdk.KEY_Down, Gdk.KEY_Left):
            rating = self.props.rating - 1
        else:
            return

        rating = max(0, rating)

        # Prevents unsetting rating if maximum is reached
        if rating == self.props.rating:
            return

        self.props.rating = rating

    def on_rating_update(self, event_type, sender, data):
        """
        Updates the rating from the current track
        """
        if self._player.current is not None:
            self._rating = self._player.current.get_rating()
            self._image.set_from_pixbuf(
                icons.MANAGER.pixbuf_from_rating(self._rating).pixbuf
            )
            self.set_sensitive(True)
        else:
            self.set_sensitive(False)


class RatingMenuItem(Gtk.MenuItem):
    """
    A menuitem containing a rating widget
    """

    __gproperties__ = {
        'rating': (
            GObject.TYPE_INT,
            'rating',
            'The selected rating',
            0,  # Minimum
            65535,  # Maximum
            0,  # Default
            GObject.ParamFlags.READWRITE,
        )
    }
    __gsignals__ = {
        'rating-changed': (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_INT,))
    }

    def __init__(self, rating=0, player=None):
        """
        :param rating: the optional initial rating
        :type rating: int
        :param player: If not None, this rating widget will automatically
                       update to reflect the rating of the current song
        :type player: xl.player.ExailePlayer
        """
        Gtk.MenuItem.__init__(self)

        box = Gtk.Box(spacing=6)
        box.pack_start(Gtk.Label(label=_('Rating:')), False, False, 0)
        self.rating_widget = RatingWidget(rating, player)
        box.pack_start(self.rating_widget, False, False, 0)

        self.add(box)

        self.rating_widget.connect('rating-changed', self.on_rating_changed)

    def do_get_property(self, property):
        """
        Getter for custom properties
        """
        if property.name == 'rating':
            return self.rating_widget.props.rating
        else:
            raise AttributeError('unknown property %s' % property.name)

    def do_set_property(self, property, value):
        """
        Setter for custom properties
        """
        if property.name == 'rating':
            self.rating_widget.props.rating = value
        else:
            raise AttributeError('unknown property %s' % property.name)

    def do_motion_notify_event(self, event):
        """
        Forwards the event to the rating widget
        """
        allocation = self.rating_widget.get_allocation()

        if allocation.x < event.x < allocation.x + allocation.width:
            x, y = self.translate_coordinates(
                self.rating_widget, int(event.x), int(event.y)
            )
            event.x, event.y = float(x), float(y)
            self.rating_widget.emit('motion-notify-event', event.copy())

    def do_leave_notify_event(self, event):
        """
        Forwards the event to the rating widget
        """
        self.rating_widget.emit('leave-notify-event', event.copy())

    def do_button_release_event(self, event):
        """
        Forwards the event to the rating widget
        """
        allocation = self.rating_widget.get_allocation()

        if allocation.x < event.x < allocation.x + allocation.width:
            x, y = self.translate_coordinates(
                self.rating_widget, int(event.x), int(event.y)
            )
            event.x, event.y = float(x), float(y)
            self.rating_widget.emit('button-release-event', event.copy())

    def on_rating_changed(self, widget, rating):
        """
        Forwards the event
        """
        self.emit('rating-changed', rating)


class RatingCellRenderer(Gtk.CellRendererPixbuf):
    """
    A cell renderer drawing rating images
    and allowing for selection of ratings
    """

    __gproperties__ = {
        'rating': (
            GObject.TYPE_INT,
            'Rating',
            'The selected rating',
            0,  # Minimum
            65535,  # Maximum
            0,  # Default
            GObject.ParamFlags.READWRITE,
        )
    }
    __gsignals__ = {
        'rating-changed': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (GObject.TYPE_PYOBJECT, GObject.TYPE_INT),
        )
    }

    def __init__(self):
        Gtk.CellRendererPixbuf.__init__(self)
        self.props.mode = Gtk.CellRendererMode.ACTIVATABLE
        self.props.xalign = 0

        self.rating = 0
        self.size_ratio = 1

    def do_get_property(self, property):
        """
        Getter for GObject properties
        """
        if property.name == 'rating':
            return self.rating
        else:
            raise AttributeError('unknown property %s' % property.name)

    def do_set_property(self, property, value):
        """
        Setter for GObject properties
        """
        if property.name == 'rating':
            self.rating = value
            self.props.pixbuf = icons.MANAGER.pixbuf_from_rating(
                self.rating, self.size_ratio
            ).pixbuf
        else:
            raise AttributeError('unknown property %s' % property.name)

    def do_activate(self, event, widget, path, background_area, cell_area, flags):
        """
        Checks if a button press event did occur
        within the clickable rating image area
        """
        if event is None:  # Keyboard activation
            return

        # Locate click area at zero
        click_area = Gdk.Rectangle()
        click_area.x = 0
        click_area.y = self.props.ypad
        click_area.width = self.props.pixbuf.get_width()
        click_area.height = self.props.pixbuf.get_height()

        # Move event location relative to zero
        event.x -= cell_area.x + click_area.x
        event.y -= cell_area.y + click_area.y

        if 0 <= event.x <= click_area.width:
            fraction = event.x / click_area.width
            maximum = settings.get_option('rating/maximum', 5)
            rating = fraction * maximum + 1
            self.emit('rating-changed', (int(path),), rating)


'''
    No longer needed in GTK3?

    def do_render(self, context, widget, background_area,
                  cell_area, flags):
        """
            Renders the rating images
            (Overridden since Gtk.CellRendererPixbuf
             fails at vertical padding)
        """
        cell_area.width *= self.props.xalign
        cell_area.height *= self.props.yalign

        pixbuf_width = self.props.pixbuf.get_width() * self.props.xalign
        pixbuf_height = self.props.pixbuf.get_height() * self.props.yalign

        x = cell_area.x + cell_area.width - pixbuf_width
        y = cell_area.y + cell_area.height - pixbuf_height

        context.set_source_pixbuf(self.props.pixbuf, x, y)
        context.paint()
'''



# moodbar - Replace Exaile's seekbar with a moodbar
# Copyright (C) 2015, 2019  Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import collections

import gi

gi.require_version('PangoCairo', '1.0')

from gi.repository import Gtk, Pango, PangoCairo

from . import painter

Extents = collections.namedtuple(
    'Extents', 'x_bearing y_bearing width height x_advance y_advance'
)


class Moodbar(Gtk.DrawingArea):
    POS_SIZE = 0.4  # Size of seek notch, in fraction of bar height
    POS_LINESIZE = 2

    def __init__(self):
        super(Moodbar, self).__init__()
        # TODO: Handle screen-changed.
        # See https://developer.gnome.org/gtk3/stable/GtkWidget.html#gtk-widget-create-pango-layout
        self.pango_layout = self.create_pango_layout()
        self.data = self.surf = self.text = self.text_extents = self.tint = None
        self.painter = None
        self._use_waveform = False
        self._seek_position = None

    def set_mood(self, data):
        """
        :param data: Mood data, or None to not show any
        :type data: Optional[bytes]
        :return: Whether the mood data is successfully set
        :rtype: bool
        """
        self.data = data
        if data:
            self.surf = self._paint(data)
            self._invalidate()
            return bool(self.surf)
        else:
            self.surf = None
            self._invalidate()
            return True

    @property
    def seek_position(self):
        return self._seek_position

    @seek_position.setter
    def seek_position(self, pos):
        """
        :param pos: Seek position, between 0 and 1 inclusive, or None to hide
        :type pos: Optional[float]
        """
        old_pos = self._seek_position
        if pos != old_pos:
            self._seek_position = pos
            self._invalidate()

    def set_text(self, text):
        """Set the text in the middle of the moodbar.

        :type text: Optional[str]
        """
        old_text = self.text
        if text != old_text:
            self.text = text
            if text:
                layout = self.pango_layout
                layout.set_text(text, -1)
                self.text_extents = layout.get_extents()[1]
            self._invalidate()

    def set_tint(self, tint):
        """Add a color layer to the whole moodbar.

        :param tint: The color tint, or None to disable
        :type tint: Optional[Gdk.ARGB]
        """
        old_tint = self.tint
        if tint != old_tint:
            self.tint = tint
            self._invalidate()

    def set_use_waveform(self, use_waveform):
        """Set whether to paint using the waveform painter (or the normal one).

        :type use_waveform: bool
        """
        old_use_waveform = self._use_waveform
        if use_waveform != old_use_waveform:
            self._use_waveform = use_waveform
            self.painter = None
            data = self.data
            if data:
                self.surf = self._paint(data)
                self._invalidate()

    def _paint(self, *args, **kwargs):
        p = self.painter
        if not p:
            self.painter = p = (
                painter.WaveformPainter()
                if self._use_waveform
                else painter.NormalPainter()
            )
        return p.paint(*args, **kwargs)

    def _invalidate(self):
        alloc = self.get_allocation()
        self.queue_draw_area(0, 0, alloc.width, alloc.height)

    def do_draw(self, cr):
        """Handler for the 'draw' signal.

        :type cr: cairo.Context
        :rtype: bool
        """
        alloc = self.get_allocation()
        width, height = alloc.width, alloc.height

        # Mood
        if self.surf:
            cr.save()
            cr.scale(width / self.surf.get_width(), height / self.surf.get_height())
            cr.set_source_surface(self.surf, 0, 0)
            cr.paint()
            cr.restore()

        # Text
        text = self.text
        if text:
            cr.save()
            # TODO: Do we need PangoCairo.update_layout anywhere?
            ext = self.text_extents
            scale = Pango.SCALE
            tx = int((width - ext.width / scale) / 2 - ext.x / scale)
            ty = int((height - ext.height / scale) / 2 - ext.y / scale)
            cr.move_to(tx, ty)
            cr.set_line_width(4)
            PangoCairo.layout_path(cr, self.pango_layout)
            cr.set_source_rgba(1, 1, 1, 0.8)
            cr.stroke_preserve()
            cr.set_source_rgb(0, 0, 0)
            cr.fill()
            cr.restore()

        # Seek position
        pos = self._seek_position
        if pos is not None:
            cr.save()
            x = pos * width
            y = height * self.POS_SIZE
            xd = y
            linesize = self.POS_LINESIZE
            top = linesize / 2 - 0.5
            # Triangle
            cr.move_to(x, y)
            cr.line_to(x - xd, top)
            cr.line_to(x + xd, top)
            cr.close_path()
            # White fill
            cr.set_source_rgb(1, 1, 1)
            cr.fill_preserve()
            # Black stroke
            cr.set_source_rgb(0, 0, 0)
            cr.set_line_width(linesize)
            cr.stroke()
            cr.restore()

        # Border
        cr.save()
        cr.rectangle(0, 0, width, height)
        if self.surf or pos is not None:
            cr.set_source_rgb(0, 0, 0)
        else:
            cr.set_source_rgb(0.63, 0.63, 0.63)
        cr.stroke()
        cr.restore()

        # Tint
        tint = self.tint
        if tint is not None:
            cr.save()
            cr.set_source_rgba(tint.red, tint.green, tint.blue, tint.alpha)
            cr.paint()
            cr.restore()

        return False


# vi: et sts=4 sw=4 tw=99



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import logging

from gi.repository import Gtk

from xl import event
from xlgui.guiutil import GtkTemplate

logger = logging.getLogger(__name__)


@GtkTemplate('ui', 'device_manager.ui')
class ManagerDialog(Gtk.Window):
    """
    the device manager dialog
    """

    __gtype_name__ = 'DeviceManager'

    tree_devices, model = GtkTemplate.Child.widgets(2)

    def __init__(self, parent, main):
        Gtk.Window.__init__(self)
        self.init_template()

        self.main = main
        self.device_manager = self.main.exaile.devices
        self.set_transient_for(parent)

        # GtkListStore self.model: first column (PyObject) should really be of
        # type devices.Device, but that doesn't work with GtkBuilder.

        self.populate_tree()
        event.add_ui_callback(self.populate_tree, 'device_added')
        event.add_ui_callback(self.populate_tree, 'device_removed')

    def populate_tree(self, *args):
        self.model.clear()
        for d in self.device_manager.get_devices():
            self.model.append([d, None, d.get_name(), d.__class__.__name__])

    def _get_selected_devices(self):
        sel = self.tree_devices.get_selection()
        model, paths = sel.get_selected_rows()
        devices = []
        for path in paths:
            iter = self.model.get_iter(path)
            device = self.model.get_value(iter, 0)
            devices.append(device)

        return devices

    @GtkTemplate.Callback
    def on_btn_connect_clicked(self, *args):
        devices = self._get_selected_devices()

        for d in devices:
            d.connect()

    @GtkTemplate.Callback
    def on_btn_disconnect_clicked(self, *args):
        devices = self._get_selected_devices()

        for d in devices:
            d.disconnect()

    @GtkTemplate.Callback
    def on_btn_close_clicked(self, *args):
        self.hide()
        self.destroy()

    def run(self):
        self.show_all()



'''
    Provided so that we can build docs without needing to have any
    dependencies installed (such as for RTFD).

    Derived from https://read-the-docs.readthedocs.org/en/latest/faq.html
'''

import sys


# from python's unittest.mock, Python license
def _is_magic(name):
    return '__%s__' % name[2:-2] == name


class MockGiModule:
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return MockGiModule()

    @classmethod
    def __getattr__(cls, name):
        if _is_magic(name):
            raise AttributeError(name)
        elif name in 'GObject':
            # This corresponds to GObject class in the GI (GObject)
            # module - we need to return type instead of an instance!
            #
            # Note: we need to do this for each class that comes from
            # a GI module AND is involved in multiple-inheritance in our
            # codebase in order to avoid "TypeError: metaclass conflict"
            # errors.
            return MockGiModule
        else:
            return MockGiModule()

    # glib mocks
    @classmethod
    def get_user_data_dir(cls):
        return '/tmp'

    @classmethod
    def get_user_config_dir(cls):
        return '/tmp'

    @classmethod
    def get_user_cache_dir(cls):
        return '/tmp'

    # gtk/gdk mocks
    class ModifierType:
        SHIFT_MASK = 0

    @classmethod
    def accelerator_parse(cls, *args):
        return [0, 0]

class Mock:
    __all__ = []

    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return Mock()

    @classmethod
    def __getattr__(cls, name):
        if _is_magic(name):
            raise AttributeError(name)
        elif name in ('Gio', 'GLib', 'GObject', 'Gst', 'Gtk', 'Gdk'):
            # These are reached via 'from gi.repository import x' and
            # correspond to GI sub-modules - need to return an instance
            # of our mock GI module
            return MockGiModule()
        else:
            return Mock()

MOCK_MODULES = [
    'cairo',

    'dbus',
    'dbus.service',

    'gi',
    'gi.repository',
    'gi.repository.Gio',
    'gi.repository.GLib',
    'gi.repository.GObject',
    'gi.repository.Gst',
    'gi.repository.Gtk',

    'mutagen',
    'mutagen.apev2',
    'mutagen.ogg',
    'mutagen.flac',
]


def import_fake_modules():
    for mod_name in MOCK_MODULES:
        sys.modules[mod_name] = Mock()


def fake_xl_settings():
    # player hack
    import xl.settings

    def option_hack(name, default):
        if name == 'player/engine':
            return 'rtfd_hack'
        else:
            return orig_get_option(name, default)

    orig_get_option = xl.settings.get_option
    xl.settings.get_option = option_hack



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from collections import OrderedDict
import datetime
import io
import os
import string
import re
from typing import Iterable, Sequence

from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango

from xl.nls import gettext as _
from xl.metadata import CoverImage
from xl import common, settings, trax, xdg

from xlgui.widgets import dialogs
from xlgui.guiutil import GtkTemplate
from xlgui import guiutil
from xl.metadata.tags import tag_data, get_default_tagdata

import logging

logger = logging.getLogger(__name__)


class TrackPropertiesDialog(GObject.Object):
    def __init__(
        self,
        parent: Gtk.Window,
        tracks: Sequence[trax.Track],
        current_position: int = 0,
    ):
        """
        :param parent: the parent window for modal operation
        :param tracks: the tracks to process
        :param current_position: the position of the currently
            selected track in the list
        """
        GObject.Object.__init__(self)

        self.builder = guiutil.get_builder(
            xdg.get_data_path('ui', 'trackproperties_dialog.ui')
        )
        self.builder.connect_signals(self)
        self.dialog = self.builder.get_object('TrackPropertiesDialog')
        self.dialog.set_transient_for(parent)

        self.__default_attributes = Pango.AttrList()
        self.__changed_attributes = Pango.AttrList()

        self.message = dialogs.MessageBar(
            parent=self.builder.get_object('main_content'),
            buttons=Gtk.ButtonsType.CLOSE,
        )

        self.remove_tag_button = self.builder.get_object('remove_tag_button')
        self.cur_track_label = self.builder.get_object('current_track_label')
        self.apply_button = self.builder.get_object('apply_button')
        self.prev_button = self.builder.get_object('prev_track_button')
        self.next_button = self.builder.get_object('next_track_button')

        self.tags_grid = self.builder.get_object('tags_grid')
        self.properties_grid = self.builder.get_object('properties_grid')
        self.rows = []

        self.new_tag_combo = self.builder.get_object('new_tag_combo')
        self.new_tag_combo_list = Gtk.ListStore(str, str)
        for tag, tag_info in tag_data.items():
            if tag_info is not None and tag_info.editable:
                self.new_tag_combo_list.append((tag, tag_info.translated_name))
        self.new_tag_combo_list.set_sort_column_id(1, Gtk.SortType.ASCENDING)
        self.new_tag_combo.set_model(self.new_tag_combo_list)
        self.add_tag_button = self.builder.get_object('add_tag_button')
        self.add_tag_button.set_sensitive(False)

        # Show these tags for all tracks, no matter what
        def_tags = [
            'tracknumber',
            'title',
            'artist',
            'albumartist',
            'album',
            'discnumber',
            'date',
            'genre',
            'cover',
            'comment',
            '__startoffset',
            '__stopoffset',
            'lyrics',
        ]

        self.def_tags = OrderedDict([(tag, tag_data[tag]) for tag in def_tags])

        # Store the tracks and a working copy
        self.tracks = tracks
        self.trackdata = self._tags_copy(tracks)
        self.trackdata_original = self._tags_copy(tracks)
        self.current_position = current_position

        self._build_from_track(self.current_position)

        self._setup_position()
        self.dialog.show()

        self.rows[0].field.grab_focus()

    def _get_field_widget(self, tag_info, ab):
        tag_type = tag_info.type

        if tag_type == 'int':
            field = TagNumField(tag_info.min, tag_info.max, all_button=ab)
        elif tag_type == 'dblnum':
            field = TagDblNumField(tag_info.min, tag_info.max, all_button=ab)
        elif tag_type == 'time':
            field = TagNumField(tag_info.min, tag_info.max, all_button=ab)
        elif tag_type == 'image':
            field = TagImageField()
        elif tag_type == 'multiline':
            field = TagTextField(all_button=ab)
        else:
            field = TagField(all_button=ab)

        return field

    def _tags_copy(self, tracks: Iterable[trax.Track]):
        l = []
        for track in tracks:
            t = {}
            for tag, tag_info in self.def_tags.items():
                if tag_info.use_disk:
                    tagval = track.get_tag_disk(tag)
                else:
                    tagval = track.get_tag_raw(tag)

                if tagval:
                    if isinstance(tagval, list):
                        t[tag] = tagval[:]
                    else:
                        t[tag] = [tagval]
                else:
                    t[tag] = ['']

                if tag_info.type == 'dblnum':
                    for i, entry in enumerate(t[tag]):
                        if len(entry.split('/')) < 2:
                            t[tag][i] += '/0'

            for tag in track.list_tags():
                if tag not in self.def_tags:
                    tag_info = tag_data.get(tag)
                    if not tag_info or not tag_info.use_disk:
                        tagval = track.get_tag_raw(tag)
                    else:
                        tagval = track.get_tag_disk(tag)
                    if isinstance(tagval, list):
                        t[tag] = tagval[:]
                    else:
                        t[tag] = [tagval]

            l.append(t)

        return l

    def _write_tag(self, track, tag, value):
        tag_info = tag_data.get(tag)
        if not tag_info or not tag_info.use_disk:
            track.set_tag_raw(tag, value)
        else:
            track.set_tag_disk(tag, value)

    def _tags_write(self, data):
        errors = []
        dialog = SavingProgressWindow(self.dialog, len(data))
        for n, trackdata in data:
            track = self.tracks[n]
            poplist = []

            try:
                for tag in trackdata:
                    if not tag.startswith("__"):
                        if tag in ("tracknumber", "discnumber") and trackdata[tag] == [
                            "0/0"
                        ]:
                            poplist.append(tag)
                            continue
                        self._write_tag(track, tag, trackdata[tag])
                    elif tag in ('__startoffset', '__stopoffset'):
                        try:
                            offset = int(trackdata[tag][0])
                        except ValueError:
                            poplist.append(tag)
                        else:
                            track.set_tag_raw(tag, offset)

                # In case a tag has been removed..
                for tag in track.list_tags():
                    if tag in tag_data:
                        if tag_data[tag] is not None:
                            try:
                                trackdata[tag]
                            except KeyError:
                                poplist.append(tag)
                    else:
                        try:
                            trackdata[tag]
                        except KeyError:
                            poplist.append(tag)

                for tag in poplist:
                    self._write_tag(track, tag, None)

                if not track.write_tags():
                    errors.append(track.get_loc_for_io())
            except Exception:
                logger.warning("Error saving track", exc_info=True)
                errors.append(track.get_loc_for_io())

            trax.track._CACHER.remove(track)
            dialog.step()
        dialog.destroy()

        if len(errors) > 0:
            self.message.clear_buttons()
            self.message.add_button(Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)
            self.message.show_error(
                _('Writing of tags failed'),
                _(
                    'Tags could not be written to the following files:\n' '{files}'
                ).format(files='\n'.join(errors)),
            )

    def _build_from_track(self, position):
        self._clear_grids()

        self.rows = []

        # Previous, next and current track label
        self.prev_button.set_sensitive(True)
        self.next_button.set_sensitive(True)

        if position == 0:
            self.prev_button.set_sensitive(False)

        if position == (len(self.trackdata) - 1):
            self.next_button.set_sensitive(False)

        self.cur_track_label.set_text(
            _("Editing track %(current)d of %(total)d")
            % {'current': self.current_position + 1, 'total': len(self.tracks)}
        )

        trackdata = self.trackdata[position]

        if len(self.trackdata) == 1:
            ab = False
        else:
            ab = True

        for tag, tag_info in self.def_tags.items():
            for i, entry in enumerate(trackdata.get(tag, [''])):
                field = self._get_field_widget(tag_info, ab)

                row = TagRow(self, self.tags_grid, field, tag, entry, i)
                self.rows.append(row)

                try:
                    if (
                        self.trackdata[self.current_position][tag]
                        != self.trackdata_original[self.current_position][tag]
                    ):
                        row.label.set_attributes(self.__changed_attributes)
                except KeyError:
                    row.label.set_attributes(self.__changed_attributes)

        for tag in trackdata:
            if tag in self.def_tags:
                continue

            try:
                tag_info = tag_data[tag]
            except KeyError:
                tag_info = get_default_tagdata(tag)

            if tag_info is None:
                continue

            for i, entry in enumerate(trackdata[tag]):
                if tag_info.editable:
                    field = self._get_field_widget(tag_info, ab)
                    self.rows.append(TagRow(self, self.tags_grid, field, tag, entry, i))
                else:
                    field = PropertyField(tag_info.type)
                    self.rows.append(
                        TagRow(self, self.properties_grid, field, tag, entry, i)
                    )

        self._check_for_changes()
        self._build_grids_from_rows()

    def _build_grids_from_rows(self):
        self._clear_grids()
        grids = [self.tags_grid, self.properties_grid]
        cur_row = {grids[0]: 0, grids[1]: 0}

        for row in self.rows:
            row.grid.insert_row(cur_row[row.grid] + 1)
            row.grid.attach(Gtk.Separator(), 0, cur_row[row.grid], 2, 1)
            cur_row[row.grid] += 1

            row.grid.insert_row(cur_row[row.grid] + 1)
            row.grid.attach(row.label, 0, cur_row[row.grid], 1, 1)
            row.grid.attach(row.field, 1, cur_row[row.grid], 1, 1)
            cur_row[row.grid] += 1

        for grid in grids:
            grid.show_all()

        self.remove_tag_button.toggled()

    def _clear_grids(self):
        """Careful, we need to delete exactly as many rows as we inserted"""
        grids = [self.tags_grid, self.properties_grid]

        for grid in grids:
            row_count = len(grid.get_children())
            for child in grid.get_children():
                grid.remove(child)

            for i in range((row_count // 3) * 2, 0, -1):
                grid.remove_row(i)

    def _save_position(self):
        width, height = self.dialog.get_size()
        if [width, height] != [
            settings.get_option('gui/trackprop_' + key, -1)
            for key in ['width', 'height']
        ]:
            settings.set_option('gui/trackprop_height', height, save=False)
            settings.set_option('gui/trackprop_width', width, save=False)
        x, y = self.dialog.get_position()
        if [x, y] != [
            settings.get_option('gui/trackprop_' + key, -1) for key in ['x', 'y']
        ]:
            settings.set_option('gui/trackprop_x', x, save=False)
            settings.set_option('gui/trackprop_y', y, save=False)

    def _setup_position(self):
        width = settings.get_option('gui/trackprop_width', 600)
        height = settings.get_option('gui/trackprop_height', 350)
        x = settings.get_option('gui/trackprop_x', 100)
        y = settings.get_option('gui/trackprop_y', 100)

        self.dialog.move(x, y)
        self.dialog.resize(width, height)

    def on_apply_button_clicked(self, w):
        modified = []
        for n, trackdata in enumerate(self.trackdata):
            if trackdata != self.trackdata_original[n]:
                modified.append((n, trackdata))

        if modified:
            if len(modified) != 1:
                dialog = Gtk.MessageDialog(
                    buttons=Gtk.ButtonsType.YES_NO,
                    message_type=Gtk.MessageType.QUESTION,
                    modal=True,
                    text=_('Are you sure you want to apply the changes to all tracks?'),
                    transient_for=self.dialog,
                )
                response = dialog.run()
                dialog.destroy()

                if response != Gtk.ResponseType.YES:
                    return

            self._tags_write(modified)

            del self.trackdata
            del self.trackdata_original
            self.trackdata = self._tags_copy(self.tracks)
            self.trackdata_original = self._tags_copy(self.tracks)

            self.apply_button.set_sensitive(False)
            for row in self.rows:
                if row.multi_id == 0:
                    row.label.set_attributes(self.__default_attributes)

        # Hide close confirmation if necessary
        if self.message.get_message_type() == Gtk.MessageType.QUESTION:
            self.message.hide()

    def _check_for_save(self):
        if self.trackdata != self.trackdata_original:

            def on_response(message, response):
                """
                Applies changes before closing if requested
                """
                if response == Gtk.ResponseType.APPLY:
                    self.apply_button.clicked()

                self._save_position()
                self.dialog.destroy()

            self.message.connect('response', on_response)
            self.message.clear_buttons()
            self.message.add_button(Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)
            self.message.add_button(Gtk.STOCK_APPLY, Gtk.ResponseType.APPLY)
            self.message.show_question(
                _('Apply changes before closing?'),
                _('Your changes will be lost if you do not apply them now.'),
            )
        else:
            self._save_position()
            self.dialog.destroy()

    def on_close_button_clicked(self, w):
        self._check_for_save()

    def on_delete_event(self, widget, data=None):
        self._check_for_save()
        return True

    def on_prev_track_button_clicked(self, widget):
        self.current_position -= 1
        self._build_from_track(self.current_position)

    def on_next_track_button_clicked(self, widget):
        self.current_position += 1
        self._build_from_track(self.current_position)

    def on_title_case_button_clicked(self, w):
        for row in self.rows:
            if isinstance(row.field, TagField) or isinstance(row.field, TagTextField):
                val = row.field.get_value()
                val = string.capwords(val, ' ')
                row.field.set_value(val)

        self._check_for_changes()

    def on_new_tag_entry_changed(self, entry):
        """
        Enables or disables the button for adding tags,
        effectively preventing empty tag names
        """
        self.add_tag_button.set_sensitive(len(entry.get_text()) > 0)

    def on_add_tag_button_clicked(self, w):
        tag = None

        index = self.new_tag_combo.get_active()

        if index != -1:
            tag = self.new_tag_combo_list[index][0]
        else:
            tag = self.new_tag_combo.get_child().get_text()

        if not tag:
            return

        trackdata = self.trackdata[self.current_position]

        try:
            trackdata[tag].append('')
        except KeyError:
            trackdata[tag] = ['']

        self._build_from_track(self.current_position)

    def on_remove_tag_button_toggled(self, widget):
        for row in self.rows:
            row.set_remove_mode(widget.get_active())

    def _check_for_changes(self):
        apply_flag = False
        for i, trackdata in enumerate(self.trackdata):
            for tag in trackdata:
                try:
                    if trackdata[tag] != self.trackdata_original[i][tag]:
                        apply_flag = True
                except KeyError:
                    apply_flag = True

            if len(trackdata) != len(self.trackdata_original[i]):
                apply_flag = True

        if apply_flag:
            if not self.apply_button.get_property("sensitive"):
                self.apply_button.set_sensitive(True)
        else:
            self.apply_button.set_sensitive(False)

    def update_tag(self, widget, tag, multi_id, val):
        trackdata = self.trackdata[self.current_position]
        original_trackdata = self.trackdata_original[self.current_position]
        trackdata[tag][multi_id] = val()

        for row in self.rows:
            if row.tag == tag and row.multi_id == 0:
                try:
                    if trackdata[tag] != original_trackdata[tag]:
                        row.label.set_attributes(self.__changed_attributes)
                    else:
                        row.label.set_attributes(self.__default_attributes)
                except KeyError:
                    row.label.set_attributes(self.__changed_attributes)

            if row.tag == tag and row.multi_id == multi_id:
                all_vals = []
                for trackdata in self.trackdata:
                    try:
                        all_vals.append(trackdata[tag][multi_id])
                    except (KeyError, IndexError):
                        all_vals.append('')

                row.field.set_value(val(), all_vals, doupdate=False)

        self._check_for_changes()

    def apply_all(self, field, multi_id, val, split_num=0):
        value = val()

        if tag_data[field].type == 'dblnum':
            original_values = value.split("/")
            for i, trackdata in enumerate(self.trackdata):
                values = trackdata[field][multi_id].split("/")
                values[split_num] = original_values[split_num]
                try:
                    trackdata[field][multi_id] = values[0] + "/" + values[1]
                except KeyError:
                    trackdata[field] = [values[0] + "/" + values[1]]
                except IndexError:
                    trackdata[field].append(values[0] + "/" + values[1])

        else:
            for i, trackdata in enumerate(self.trackdata):
                try:
                    trackdata[field][multi_id] = value
                except KeyError:
                    trackdata[field] = [value]
                except IndexError:
                    trackdata[field].append(value)

        self._check_for_changes()

    def remove_row(self, w, tag, multi_id):
        for row in self.rows:
            if row.tag == tag and row.multi_id == multi_id:
                self.trackdata[self.current_position][tag].pop(multi_id)
                if len(self.trackdata[self.current_position][tag]) == 0:
                    self.trackdata[self.current_position].pop(tag)

        self._build_from_track(self.current_position)

    def run(self):
        return self.dialog.run()

    def hide(self):
        self.dialog.hide()


class TagRow:
    def __init__(self, parent, parent_grid, field, tag_name, value, multi_id):
        self.parent = parent
        self.grid = parent_grid
        self.tag = tag_name
        self.field = field
        self.field.register_parent_row(self)
        self.multi_id = multi_id
        all_vals = []
        for track in parent.trackdata:
            try:
                all_vals.append(track[tag_name][multi_id])
            except (KeyError, IndexError):
                all_vals.append(None)

        self.field.set_value(value, all_vals)

        try:
            name = tag_data[self.tag].translated_name
        except KeyError:
            if self.tag.startswith('__'):
                name = self.tag[2:]
            else:
                name = self.tag
            name = name.capitalize()

        self.name = name
        self.label = Gtk.Label(halign=Gtk.Align.START, margin_top=5)

        if multi_id == 0:
            # TRANSLATORS: Label for a tag on the Track Properties dialog
            self.label.set_text(_('%s:') % name)
            self.label.create_pango_context()

        self.clear_button = Gtk.Button(
            image=Gtk.Image.new_from_icon_name('edit-clear', Gtk.IconSize.BUTTON),
            relief=Gtk.ReliefStyle.NONE,
            # TRANSLATORS: Remove tag value
            tooltip_text=_("Clear"),
        )
        self.clear_button.connect("clicked", self.clear)

        if not isinstance(field, PropertyField):
            self.field.pack_start(self.clear_button, False, False, 0)

        self.field.show_all()

        # Remove mode settings
        self.remove_mode = False
        self.remove_button = Gtk.Button()
        self.remove_button.set_image(
            Gtk.Image.new_from_icon_name('list-remove', Gtk.IconSize.BUTTON)
        )
        self.remove_button.connect(
            "clicked", parent.remove_row, self.tag, self.multi_id
        )

        self.field.register_update_func(parent.update_tag)
        self.field.register_all_func(parent.apply_all)

    def set_remove_mode(self, val):
        if not self.tag.startswith('__') or self.multi_id != 0:
            if val and not self.remove_mode:
                self.field.remove(self.clear_button)
                self.field.pack_start(self.remove_button, False, False, 0)
                self.field.show_all()
                self.remove_mode = True

            if not val and self.remove_mode:
                self.field.remove(self.remove_button)
                self.remove_mode = False
                self.field.pack_start(self.clear_button, False, False, 0)
                self.field.show_all()

    def clear(self, w):
        self.field.set_value('')


class TagField(Gtk.Box):
    def __init__(self, all_button=True):
        Gtk.Box.__init__(self, homogeneous=False, spacing=5)

        # Create the widgets
        self.field = Gtk.Entry()
        self.all_func = None
        self.parent_row = None

        self.pack_start(self.field, True, True, 0)

        self.all_button = None
        if all_button:
            self.all_button = AllButton(self)
            self.pack_start(self.all_button, False, False, 0)

    def grab_focus(self):
        """
        Gives focus to the internal widget
        """
        self.field.grab_focus()

    def register_parent_row(self, parent_row):
        self.parent_row = parent_row

    def set_value(self, val, all_vals=None, doupdate=True):
        if doupdate:
            self.field.set_text(val)

        if all_vals is not None and self.all_button is not None:
            # Set the value of the all button
            self.all_button.set_active(all(val == v for v in all_vals))

    def get_value(self):
        return self.field.get_text()

    def register_update_func(self, f):
        tag = self.parent_row.tag
        multi_id = self.parent_row.multi_id
        self.field.connect("changed", f, tag, multi_id, self.get_value)

    def register_all_func(self, f):
        self.all_func = f


def dummy_scroll_handler(widget, _):
    """scroll-event handler that just disables the default handler.

    Tag field widgets should use this to prevent the user from accidentally
    modifying tags by scrolling.
    """
    GObject.signal_stop_emission_by_name(widget, 'scroll-event')
    return False


class TagTextField(Gtk.Box):
    def __init__(self, all_button=True):
        Gtk.Box.__init__(self, homogeneous=False, spacing=5)

        self.buffer = Gtk.TextBuffer()
        self.field = Gtk.TextView.new_with_buffer(self.buffer)
        self.field.set_size_request(200, 150)  # XXX
        self.field.set_wrap_mode(Gtk.WrapMode.WORD_CHAR)
        scrollwindow = Gtk.ScrolledWindow()
        scrollwindow.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        scrollwindow.set_shadow_type(Gtk.ShadowType.IN)
        scrollwindow.add(self.field)
        self.all_func = None
        self.parent_row = None

        self.pack_start(scrollwindow, True, True, 0)

        self.all_button = None
        if all_button:
            self.all_button = AllButton(self)
            self.pack_start(self.all_button, False, False, 0)

    def register_parent_row(self, parent_row):
        self.parent_row = parent_row

    def set_value(self, val, all_vals=None, doupdate=True):
        if doupdate:
            self.buffer.set_text(val)

        if all_vals is not None and self.all_button is not None:
            # Set the value of the all button
            flag = True
            for v in all_vals:
                if val != v:
                    flag = False

            if flag:
                self.all_button.set_active(True)
            else:
                self.all_button.set_active(False)

    def get_value(self):
        return self.buffer.get_text(
            self.buffer.get_start_iter(), self.buffer.get_end_iter(), True
        )

    def register_update_func(self, f):
        tag = self.parent_row.tag
        multi_id = self.parent_row.multi_id
        self.buffer.connect("changed", f, tag, multi_id, self.get_value)

    def register_all_func(self, f):
        self.all_func = f


class TagNumField(Gtk.Box):
    def __init__(self, min=0, max=10000, step=1, page=10, all_button=True):
        Gtk.Box.__init__(self, homogeneous=False, spacing=5)

        # Create the widgets
        self.field = Gtk.SpinButton()
        self.field.set_range(min, max)
        self.field.set_increments(step, page)
        self.field.connect('scroll-event', dummy_scroll_handler)
        self.all_func = None
        self.parent_row = None

        self.pack_start(self.field, True, True, 0)

        self.all_button = None
        if all_button:
            self.all_button = AllButton(self)
            self.pack_start(self.all_button, False, False, 0)

    def register_parent_row(self, parent_row):
        self.parent_row = parent_row

    def set_value(self, val, all_vals=None, doupdate=True):
        if doupdate:
            field_val = 0
            if val != '':
                try:
                    field_val = float(val)
                except ValueError:
                    digits = re.search(r'\d+\.?\d*', val)
                    if digits:
                        field_val = float(digits.group())
            self.field.set_value(field_val)

        if all_vals is not None and self.all_button is not None:
            # Set the value of the all button
            flag = True
            for v in all_vals:
                if val != v:
                    flag = False

            if flag:
                self.all_button.set_active(True)
            else:
                self.all_button.set_active(False)

    def get_value(self):
        return str(int(self.field.get_value()))

    def register_update_func(self, f):
        tag = self.parent_row.tag
        multi_id = self.parent_row.multi_id
        self.field.connect("value-changed", f, tag, multi_id, self.get_value)

    def register_all_func(self, f):
        self.all_func = f


class TagDblNumField(Gtk.Box):
    def __init__(self, min=0, max=10000, step=1, page=10, all_button=True):
        Gtk.Box.__init__(self, homogeneous=False, spacing=5)

        self.field = [Gtk.SpinButton(), Gtk.SpinButton()]
        self.all_func = None
        self.parent_row = None
        for f in self.field:
            f.set_range(min, max)
            f.set_increments(step, page)
            f.connect('scroll-event', dummy_scroll_handler)

        # TRANSLATORS: This is the 'of' between numbers in fields like
        # tracknumber, discnumber, etc. in the tagger.
        lbl = Gtk.Label(label=_('of:'))
        self.all_button = [None, None]
        if all_button:
            self.all_button = [AllButton(self), AllButton(self, 1)]

        self.pack_start(self.field[0], True, True, 0)
        if all_button and self.all_button[0] is not None:
            self.pack_start(self.all_button[0], False, False, 0)
        self.pack_start(lbl, True, True, 0)
        self.pack_start(self.field[1], True, True, 0)
        if all_button:
            self.pack_start(self.all_button[1], False, False, 0)

    def grab_focus(self):
        """
        Gives focus to the internal widget
        """
        self.field[0].grab_focus()

    def register_parent_row(self, parent_row):
        self.parent_row = parent_row

    def register_all_func(self, f):
        self.all_func = f

    def set_value(self, val, all_val=None, doupdate=True):
        if len(val.split('/')) < 2:
            val += '/'

        vals = val.split('/')

        if doupdate:
            for x in range(2):
                field_val = 0
                if vals[x] != '':
                    try:
                        field_val = float(vals[x])
                    except ValueError:
                        digits = re.search(r'\d+\.?\d*', val)
                        if digits:
                            field_val = float(digits.group())
                self.field[x].set_value(field_val)

        if all_val is not None:
            all_vals = []
            for v in all_val:
                if v is not None:
                    if len(v.split('/')) < 2:
                        v += '/'
                    all_vals.append(v.split('/'))
                else:
                    all_vals.append(None)

            # Set the value of the all button
            flags = [True, True]
            for i in range(2):
                for v in all_vals:
                    if v is None or vals[i] != v[i]:
                        flags[i] = False

                if self.all_button[i] is not None:
                    if flags[i]:
                        self.all_button[i].set_active(True)
                    else:
                        self.all_button[i].set_active(False)

    def get_value(self):
        return '%d/%d' % (self.field[0].get_value(), self.field[1].get_value())

    def register_update_func(self, f):
        tag = self.parent_row.tag
        multi_id = self.parent_row.multi_id
        self.field[0].connect("value-changed", f, tag, multi_id, self.get_value)
        self.field[1].connect("value-changed", f, tag, multi_id, self.get_value)


@GtkTemplate('ui', 'trackproperties_dialog_cover_row.ui')
class TagImageField(Gtk.Box):
    __gtype_name__ = 'TagImageField'

    (
        button,
        image,
        type_model,
        description_entry,
        type_selection,
        info_label,
    ) = GtkTemplate.Child.widgets(6)

    def __init__(self, all_button=True):
        Gtk.Box.__init__(self)
        self.init_template()

        self.parent_row = None
        self.all_func = None
        self.update_func = None
        # Prevents the update function from being called, make
        # sure you do that manually after the batch update
        self.batch_update = False

        self.pixbuf = None
        self.info = CoverImage(None, None, None, None)
        self.default_type = 3
        self.mime_info = {
            'image/jpeg': {
                # Title for display
                'title': _('JPEG image'),
                # Type and options for GDK Pixbuf saving
                'type': 'jpeg',
                'options': {'quality': '90'},
            },
            'image/png': {'title': _('PNG image'), 'type': 'png', 'options': {}},
            'image/': {
                'title': _('Image'),
                # Store unknown images as JPEG
                'type': 'jpeg',
                'options': {'quality': '90'},
            },
            # TODO: Handle linked images
            '-->': {'title': _('Linked image')},
        }

        self.button.drag_dest_set(Gtk.DestDefaults.ALL, [], Gdk.DragAction.COPY)
        self.button.drag_dest_add_uri_targets()

        self.type_selection.connect('scroll-event', dummy_scroll_handler)

        self.all_button = None
        if all_button:
            self.all_button = AllButton(self)
            self.pack_start(self.all_button, False, False, 0)

    def grab_focus(self):
        """
        Gives focus to the internal widget
        """
        self.image.grab_focus()

    def register_parent_row(self, parent_row):
        self.parent_row = parent_row

    def register_update_func(self, func):
        self.update_func = func

    def register_all_func(self, function):
        self.all_func = function

    def set_value(self, val, all_vals=None, doupdate=True):
        if doupdate:
            if val:
                loader = GdkPixbuf.PixbufLoader()

                try:
                    loader.write(val.data)
                    loader.close()
                except GLib.Error:
                    pass
                else:
                    self.batch_update = True
                    self.set_pixbuf(loader.get_pixbuf(), val.mime)

                    # some file types do not support multiple cover types
                    if val.type is not None:
                        self.type_selection.set_active(val.type)
                        self.type_selection.set_sensitive(True)
                    else:
                        self.type_selection.set_active(-1)
                        self.type_selection.set_sensitive(False)

                    if val.desc is not None:
                        self.description_entry.set_text(val.desc)
                        self.description_entry.set_sensitive(True)
                    else:
                        self.description_entry.set_text('')
                        self.description_entry.set_sensitive(False)

                    self.batch_update = False
            else:
                self.batch_update = True
                self.set_pixbuf(None)
                self.type_selection.set_active(-1)
                self.type_selection.set_sensitive(False)
                self.description_entry.set_text('')
                self.description_entry.set_sensitive(False)
                self.batch_update = False
                self.call_update_func()

        if None not in (all_vals, self.all_button):
            self.all_button.set_active(all(val == v for v in all_vals))

    def get_value(self):
        if not self.pixbuf:
            return None

        mime = self.mime_info[self.info.mime]
        # Retrieve proper image data
        writer = io.BytesIO()

        def gdk_pixbuf_save_func(buf, count, user_data):
            if writer.write(buf) == count:
                return True
            return False

        # workaround for some undocumented changes in GdkPixbuf API
        # see https://bugzilla.gnome.org/show_bug.cgi?id=670372
        # see https://github.com/mypaint/mypaint/issues/236
        try:
            save_to_callback_function = self.pixbuf.save_to_callbackv
        except AttributeError:
            save_to_callback_function = self.pixbuf.save_to_callback
        save_to_callback_function(
            gdk_pixbuf_save_func,
            None,
            mime['type'],
            list(mime['options'].keys()),  # must be a sequence (list)
            list(mime['options'].values()),  # must be a sequence (list)
        )

        # Move to the beginning of the buffer to allow read operations
        writer.seek(0)

        return self.info._replace(data=writer.read())

    def call_update_func(self):
        """
        Wrapper around the update function
        """
        if not self.update_func or self.batch_update:
            return

        self.update_func(
            self, self.parent_row.tag, self.parent_row.multi_id, self.get_value
        )

    def set_pixbuf(self, pixbuf, mime=None):
        """
        Updates the displayed cover image and info values
        """
        self.pixbuf = pixbuf

        if pixbuf is None:
            self.image.set_from_icon_name('list-add', Gtk.IconSize.DIALOG)
            self.info_label.set_markup('')
        else:
            self.image.set_from_pixbuf(
                pixbuf.scale_simple(100, 100, GdkPixbuf.InterpType.BILINEAR)
            )

            width, height = pixbuf.get_width(), pixbuf.get_height()
            if mime is None:
                # TRANSLATORS: do not translate 'width' and 'height'
                markup = _('{width}x{height} pixels').format(width=width, height=height)
            else:
                # TRANSLATORS: do not translate 'format', 'width', and 'height'
                markup = _('{format} ({width}x{height} pixels)').format(
                    format=self.mime_info.get(mime, self.mime_info['image/'])['title'],
                    width=width,
                    height=height,
                )
            self.info_label.set_markup(markup)

            self.info = self.info._replace(mime=mime)

    def _on_button_clicked(self, button):
        """
        Allows setting the cover image using a file selection dialog
        """
        dialog = dialogs.FileOperationDialog(
            title=_('Select image to set as cover'),
            parent=self.get_toplevel(),
            buttons=(
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_OK,
                Gtk.ResponseType.OK,
            ),
        )
        dialog.set_select_multiple(False)
        filefilter = Gtk.FileFilter()
        # Not using Gtk.FileFilter.add_pixbuf_formats since
        # not all image formats are supported in tags
        filefilter.set_name(_('Supported image formats'))
        filefilter.add_pattern('*.[jJ][pP][gG]')
        filefilter.add_pattern('*.[jJ][pP][eE][gG]')
        filefilter.add_pattern('*.[pP][nN][gG]')
        dialog.add_filter(filefilter)

        if dialog.run() == Gtk.ResponseType.OK:
            filename = dialog.get_filename()

            try:
                pixbuf = GdkPixbuf.Pixbuf.new_from_file(filename)
                info = GdkPixbuf.Pixbuf.get_file_info(filename)[0]
            except TypeError:
                pass
            else:
                self.batch_update = True
                self.set_pixbuf(pixbuf, info.get_mime_types()[0])
                self.type_selection.set_active(self.default_type)
                self.type_selection.set_sensitive(True)
                self.description_entry.set_text(
                    os.path.basename(filename).rsplit('.', 1)[0]
                )
                self.description_entry.set_sensitive(True)
                self.batch_update = False
                self.call_update_func()

        dialog.destroy()

    def _on_button_drag_data_received(
        self, widget, context, x, y, selection, info, time
    ):
        """
        Allows setting the cover image via drag and drop
        """
        if selection.target.name() == 'text/uri-list':
            filename = Gio.File.new_for_uri(selection.get_uris()[0]).get_path()

            try:
                pixbuf = GdkPixbuf.Pixbuf.new_from_file(filename)
                info = GdkPixbuf.Pixbuf.get_file_info(filename)[0]
            except TypeError:
                pass
            else:
                self.batch_update = True
                self.set_pixbuf(pixbuf, info['mime_types'][0])
                self.type_selection.set_active(self.default_type)
                self.description_entry.set_sensitive(True)
                self.description_entry.set_text(
                    os.path.basename(filename).rsplit('.', 1)[0]
                )
                self.description_entry.set_sensitive(True)
                self.batch_update = False
                self.call_update_func()

    def _on_type_selection_changed(self, combobox):
        """
        Notifies about changes in the cover type
        """
        self.info = self.info._replace(type=self.type_model[combobox.get_active()][0])
        self.call_update_func()

    def _on_description_entry_changed(self, entry):
        """
        Notifies about changes in the cover description
        """
        self.info = self.info._replace(desc=entry.get_text())
        self.call_update_func()


class PropertyField(Gtk.Box):
    def __init__(self, property_type='text'):
        Gtk.Box.__init__(self, homogeneous=False, spacing=5)

        # Informs of special formatting required
        self.property_type = property_type
        self.field = Gtk.Entry()
        self.field.set_editable(False)
        self.pack_start(self.field, True, True, 0)
        self.set_hexpand(True)
        self.parent_row = None

        if self.property_type == 'location':
            self.folder_button = Gtk.Button()
            self.folder_button.set_tooltip_text(_('Open Directory'))
            self.folder_button.set_image(
                Gtk.Image.new_from_icon_name('folder-open', Gtk.IconSize.BUTTON)
            )
            self.pack_start(self.folder_button, False, False, 0)
            self.folder_button.connect("clicked", self.folder_button_clicked)

    def register_parent_row(self, parent_row):
        self.parent_row = parent_row

    def set_value(self, val, all_vals=None, doupdate=True):
        if self.property_type == 'bitrate':
            try:
                val = str(float(val) / 1000.0) + ' kbps'
            except (TypeError, ValueError):
                pass
        elif self.property_type == 'timestamp':
            d = datetime.datetime.fromtimestamp(val)
            val = d.strftime("%x %X")
        elif self.property_type == 'time':
            val = "%(m)d:%(s)02d" % {'m': val // 60, 's': val % 60}
        elif self.property_type == 'location':
            f = Gio.File.new_for_uri(val)
            val = f.get_parse_name()

            if not f.get_path():
                # Sanitize URLs of remote locations
                val = common.sanitize_url(val)
                # Disable folder button for non-browsable locations
                self.folder_button.set_sensitive(False)
        else:
            val = str(val)

        if doupdate:
            self.field.set_text(val)
            self.field.set_tooltip_text(val)

    def folder_button_clicked(self, w):
        common.open_file_directory(self.field.get_text())

    def register_update_func(self, f):
        pass

    def register_all_func(self, f):
        pass


class AllButton(Gtk.ToggleButton):
    def __init__(self, parent_field, id_num=0):
        Gtk.ToggleButton.__init__(self)
        self.set_tooltip_text(_("Apply current value to all tracks"))
        self.set_relief(Gtk.ReliefStyle.NONE)
        self.connect("toggled", self.set_all_mode)
        self.field = parent_field
        self.id_num = id_num
        im = Gtk.Image.new_from_icon_name('edit-copy', Gtk.IconSize.BUTTON)
        self.set_image(im)
        self.set_active(True)
        self.set_active(False)

    def set_all_mode(self, w=None, do_apply=True):
        if self.get_active and do_apply and self.field.parent_row:
            tag = self.field.parent_row.tag
            multi_id = self.field.parent_row.multi_id
            if self.field.all_func is not None:
                self.field.all_func(tag, multi_id, self.field.get_value, self.id_num)


class SavingProgressWindow(Gtk.Window):
    def __init__(self, parent, total, text=_("Saved %(count)s of %(total)s.")):
        Gtk.Window.__init__(self)
        self.count = 0
        self.total = total
        self.text = text

        if parent:
            self.set_transient_for(parent)
        self.set_modal(True)
        self.set_decorated(False)
        self.set_resizable(False)
        self.set_focus_on_map(False)
        self.add(Gtk.Frame())
        self.get_child().set_shadow_type(Gtk.ShadowType.OUT)
        vbox = Gtk.Box(spacing=12, orientation=Gtk.Orientation.VERTICAL)
        vbox.set_border_width(12)
        self._label = Gtk.Label()
        self._label.set_use_markup(True)
        self._label.set_markup(self.text % {'count': 0, 'total': self.total})
        vbox.pack_start(self._label, True, True, 0)
        self._progress = Gtk.ProgressBar()
        self._progress.set_size_request(300, -1)
        vbox.pack_start(self._progress, True, True, 0)

        self.get_child().add(vbox)

        self.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
        self.show_all()
        while Gtk.events_pending():
            Gtk.main_iteration()

    def step(self):
        self.count += 1
        self._progress.set_fraction(common.clamp(self.count / float(self.total), 0, 1))
        self._label.set_markup(self.text % {'count': self.count, 'total': self.total})
        while Gtk.events_pending():
            Gtk.main_iteration()


# vim: et sts=4 sw=4



from hashlib import md5
import logging
import os
import os.path

import feedparser

from gi.repository import GLib
from gi.repository import Gtk

from xl import event, common, playlist, providers
from xl import trax
from xl.nls import gettext as _
from xlgui import panel, main
from xlgui.widgets import dialogs, menu
from xl import xdg

# set up logger
logger = logging.getLogger(__name__)

PODCASTS = None
CURPATH = os.path.realpath(__file__)
BASEDIR = os.path.dirname(CURPATH)


def enable(exaile):
    feedparser.USER_AGENT = exaile.get_user_agent_string('podcasts')
    if exaile.loading:
        event.add_callback(exaile_ready, 'gui_loaded')
    else:
        exaile_ready(None, exaile, None)


def exaile_ready(event, exaile, nothing):
    global PODCASTS

    if not PODCASTS:
        PODCASTS = PodcastPanel(main.mainwindow().window)
        providers.register('main-panel', PODCASTS)


def disable(exaile):
    global PODCASTS

    if PODCASTS:
        providers.unregister('main-panel', PODCASTS)
        PODCASTS = None


class PodcastPanel(panel.Panel):
    ui_info = (os.path.join(BASEDIR, 'podcasts.ui'), 'PodcastPanel')

    def __init__(self, parent):
        panel.Panel.__init__(self, parent, 'podcasts', _('Podcasts'))
        self.podcasts = []
        self.podcast_playlists = playlist.PlaylistManager('podcast_plugin_playlists')

        self._setup_widgets()
        self._connect_events()
        self.podcast_file = os.path.join(
            xdg.get_plugin_data_dir(), 'podcasts_plugin.db'
        )
        self._load_podcasts()

    def _setup_widgets(self):
        self.model = Gtk.ListStore(str, str)
        self.tree = self.builder.get_object('podcast_tree')
        self.tree.set_model(self.model)

        text = Gtk.CellRendererText()
        self.column = Gtk.TreeViewColumn(_('Podcast'))
        self.column.pack_start(text, True)
        self.column.set_expand(True)
        self.column.set_attributes(text, text=0)
        self.tree.append_column(self.column)

        self.status = self.builder.get_object('podcast_statusbar')

        self.menu = menu.Menu(self)
        self.menu.add_simple(_('Refresh Podcast'), self._on_refresh, Gtk.STOCK_REFRESH)
        self.menu.add_simple(_('Delete'), self._on_delete, Gtk.STOCK_DELETE)
        self.menu.attach_to_widget(self.tree)

    @common.idle_add()
    def _set_status(self, message, timeout=0):
        self.status.set_text(message)

        if timeout:
            GLib.timeout_add_seconds(timeout, self._set_status, '', 0)

    def _connect_events(self):
        self.builder.connect_signals({'on_add_button_clicked': self.on_add_podcast})

        self.tree.connect('row-activated', self._on_row_activated)
        self.tree.connect('button-press-event', self._on_button_press)

    def _on_button_press(self, button, event):
        if event.triggers_context_menu():
            self.menu.popup(event)

    def _on_refresh(self, *e):
        url, title = self.get_selected_podcast()
        self._parse_podcast(url)

    def _on_delete(self, *e):
        url, title = self.get_selected_podcast()
        for item in self.podcasts:
            title, _url = item
            if _url == url:
                self.podcasts.remove(item)
                self.podcast_playlists.remove_playlist(
                    md5(url.encode('utf-8')).hexdigest()
                )
                break

        self._save_podcasts()
        self._load_podcasts()

    def on_add_podcast(self, *e):
        dialog = dialogs.TextEntryDialog(
            _('Enter the URL of the ' 'podcast to add'), _('Open Podcast')
        )
        dialog.set_transient_for(self.parent)
        dialog.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)

        result = dialog.run()
        dialog.hide()

        if result == Gtk.ResponseType.OK:
            url = dialog.get_value()
            self._parse_podcast(url, True)

    def get_selected_podcast(self):
        selection = self.tree.get_selection()
        model, iter = selection.get_selected()

        url = self.model.get_value(iter, 1)
        title = self.model.get_value(iter, 0)
        return (url, title)

    def _on_row_activated(self, *e):
        url, title = self.get_selected_podcast()

        try:
            pl = self.podcast_playlists.get_playlist(
                md5(url.encode('utf-8')).hexdigest()
            )
            self._open_podcast(pl, title)
        except ValueError:
            self._parse_podcast(url)

    @common.threaded
    def _parse_podcast(self, url, add_to_db=False):
        try:
            url = url.replace('itpc://', 'http://')

            self._set_status(_('Loading %s...') % url)
            d = feedparser.parse(url)
            entries = d['entries']

            title = d['feed']['title']

            if add_to_db:
                self._add_to_db(url, title)

            pl = playlist.Playlist(md5(url.encode('utf-8')).hexdigest())

            tracks = []
            for e in entries:
                for link in e.get('enclosures', []):
                    tr = trax.Track(link.href)
                    date = e['updated_parsed']
                    tr.set_tags(
                        artist=title,
                        title='%s: %s' % (e['title'], link.href.split('/')[-1]),
                        date="%d-%02d-%02d" % (date.tm_year, date.tm_mon, date.tm_mday),
                    )
                    tracks.append(tr)

            pl.extend(tracks)
            self._set_status('')

            self._open_podcast(pl, title)
            self.podcast_playlists.save_playlist(pl, overwrite=True)
        except Exception:
            logger.exception("Error loading podcast")
            self._set_status(_('Error loading podcast.'), 2)

    @common.idle_add()
    def _add_to_db(self, url, title):
        self.podcasts.append((title, url))
        self._save_podcasts()
        self._load_podcasts()

    @common.idle_add()
    def _open_podcast(self, pl, title):
        new_pl = playlist.Playlist(title)
        new_pl.extend(pl)
        main.get_playlist_notebook().create_tab_from_playlist(new_pl)

    @common.threaded
    def _load_podcasts(self):
        self._set_status(_("Loading Podcasts..."))
        if not os.path.exists(self.podcast_file):
            self._set_status(_("No configuration present yet"))
            return
        try:
            with open(self.podcast_file) as fp:
                lines = (line.strip() for line in fp.readlines())

            self.podcasts = []

            for line in lines:
                url, title = line.split('\t')
                self.podcasts.append((title, url))
        except (IOError, OSError):
            logger.warning('Could not open podcast file')
            self._set_status('')
            return

        self._done_loading_podcasts()

    @common.idle_add()
    def _done_loading_podcasts(self):
        self.model.clear()
        self.podcasts.sort()
        for title, url in self.podcasts:
            self.model.append([title, url])

        self._set_status('')

    def _save_podcasts(self):
        try:
            with open(self.podcast_file, 'w') as fp:
                for title, url in self.podcasts:
                    fp.write('%s\t%s\n' % (url, title))
        except (OSError, IOError):
            dialogs.error(self.parent, _('Could not save podcast file'))
            return



# Copyright (C) 2006-2010  Johannes Sasongko <sasongko@gmail.com>
# Copyright (C) 2010  Mathias Brodala <info@noctus.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import cairo

from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import GLib
from gi.repository import Gtk

from xl import covers, event, player, settings

from xlgui.guiutil import get_workarea_dimensions, pixbuf_from_data

from . import desktopcover_preferences

DESKTOPCOVER = None


class DesktopCoverPlugin:
    def __init__(self):
        self.__desktop_cover = None

    def enable(self, _exaile):
        """
        Enables the plugin
        """
        self.__migrate_anchor_setting()

    def on_gui_loaded(self):
        self.__desktop_cover = DesktopCover()

    def disable(self, _exaile):
        """
        Disables the desktop cover plugin
        """
        self.__desktop_cover.destroy()
        self.__desktop_cover = None

    @staticmethod
    def __migrate_anchor_setting():
        """
        Migrates gravity setting from the old
        integer values to the new string values
        """
        gravity = settings.get_option('plugin/desktopcover/anchor', 'topleft')
        gravity_map = DesktopCover.gravity_map

        if gravity not in gravity_map:
            # Convert to list so we can index it with the old integer
            # setting value
            gravities = list(gravity_map.keys())

            try:
                gravity = gravities[gravity]
            except (IndexError, TypeError):
                gravity = 'topleft'

            settings.set_option('plugin/desktopcover/anchor', gravity)

    def get_preferences_pane(self):
        return desktopcover_preferences


plugin_class = DesktopCoverPlugin


class DesktopCover(Gtk.Window):
    gravity_map = {
        'topleft': Gdk.Gravity.NORTH_WEST,
        'topright': Gdk.Gravity.NORTH_EAST,
        'bottomleft': Gdk.Gravity.SOUTH_WEST,
        'bottomright': Gdk.Gravity.SOUTH_EAST,
    }

    def __init__(self):
        Gtk.Window.__init__(self)

        self.image = Gtk.Image()
        self.add(self.image)
        self.image.show()

        self.set_accept_focus(False)
        self.set_app_paintable(True)
        self.set_decorated(False)
        self.set_keep_below(True)
        self.set_resizable(False)
        self.set_role("desktopcover")
        self.set_skip_pager_hint(True)
        self.set_skip_taskbar_hint(True)
        self.set_title("Exaile desktop cover")
        self.stick()

        self._fade_in_id = None
        self._fade_out_id = None
        self._cross_fade_id = None
        self._cross_fade_step = 0
        self._events = [
            'playback_track_start',
            'playback_player_end',
            'cover_set',
            'cover_removed',
        ]

        screen = self.get_screen()
        visual = screen.get_rgba_visual() or screen.get_rgb_visual()
        self.set_visual(visual)

        if player.PLAYER.current is not None:
            self.set_cover_from_track(player.PLAYER.current)
            GLib.idle_add(self.update_position)

        for e in self._events:
            if 'playback' in e:
                event.add_ui_callback(getattr(self, 'on_%s' % e), e, player.PLAYER)
            else:
                event.add_ui_callback(getattr(self, 'on_%s' % e), e)

        event.add_ui_callback(self.on_option_set, 'plugin_desktopcover_option_set')

        self.connect('draw', self.on_draw)
        self.connect('screen-changed', self.on_screen_changed)

    def destroy(self):
        """
        Cleanups
        """
        for e in self._events:
            if 'playback' in e:
                event.remove_callback(getattr(self, 'on_%s' % e), e, player.PLAYER)
            else:
                event.remove_callback(getattr(self, 'on_%s' % e), e)

        event.remove_callback(self.on_option_set, 'plugin_desktopcover_option_set')

        Gtk.Window.destroy(self)

    def set_cover_from_track(self, track):
        """
        Updates the cover image and triggers cross-fading
        """
        cover_data = covers.MANAGER.get_cover(track, set_only=True)

        if cover_data is None:
            self.hide()
            return

        if not self.props.visible:
            self.show()

        size = settings.get_option('plugin/desktopcover/size', 200)
        upscale = settings.get_option('plugin/desktopcover/override_size', False)
        pixbuf = self.image.get_pixbuf()
        next_pixbuf = pixbuf_from_data(cover_data, size=(size, size), upscale=upscale)
        fading = settings.get_option('plugin/desktopcover/fading', False)

        if fading and pixbuf is not None and self._cross_fade_id is None:
            # Prescale to allow for proper crossfading
            width, height = next_pixbuf.get_width(), next_pixbuf.get_height()
            pixbuf = pixbuf.scale_simple(width, height, GdkPixbuf.InterpType.BILINEAR)
            self.image.set_from_pixbuf(pixbuf)

            duration = settings.get_option('plugin/desktopcover/fading_duration', 50)

            self._cross_fade_id = GLib.timeout_add(
                int(duration), self.cross_fade, pixbuf, next_pixbuf, duration
            )
        else:
            self.image.set_from_pixbuf(next_pixbuf)

    def update_position(self):
        """
        Updates the position based
        on gravity and offsets
        """
        gravity = self.gravity_map[
            settings.get_option('plugin/desktopcover/anchor', 'topleft')
        ]
        cover_offset_x = settings.get_option('plugin/desktopcover/x', 0)
        cover_offset_y = settings.get_option('plugin/desktopcover/y', 0)
        allocation = self.get_allocation()
        workarea = get_workarea_dimensions()
        x, y = workarea.x, workarea.y

        if gravity in (Gdk.Gravity.NORTH_WEST, Gdk.Gravity.SOUTH_WEST):
            x += cover_offset_x
        else:
            x += workarea.width - allocation.width - cover_offset_x

        if gravity in (Gdk.Gravity.NORTH_WEST, Gdk.Gravity.NORTH_EAST):
            y += cover_offset_y
        else:
            y += workarea.height - allocation.height - cover_offset_y

        self.set_gravity(gravity)
        self.move(int(x), int(y))

    def show(self):
        """
        Override for fade-in
        """
        fading = settings.get_option('plugin/desktopcover/fading', False)

        if fading and self._fade_in_id is None:
            self.set_opacity(0)

        Gtk.Window.show(self)

        if fading and self._fade_in_id is None:
            duration = settings.get_option('plugin/desktopcover/fading_duration', 50)
            self._fade_in_id = GLib.timeout_add(int(duration), self.fade_in)

    def hide(self):
        """
        Override for fade-out
        """
        fading = settings.get_option('plugin/desktopcover/fading', False)

        if fading and self._fade_out_id is None:
            duration = settings.get_option('plugin/desktopcover/fading_duration', 50)
            self._fade_out_id = GLib.timeout_add(int(duration), self.fade_out)
        else:
            Gtk.Window.hide(self)
            self.image.set_from_pixbuf(None)

    def fade_in(self):
        """
        Increases opacity until completely opaque
        """
        opacity = self.get_opacity()

        if opacity == 1:
            self._fade_in_id = None

            return False

        self.set_opacity(opacity + 0.1)

        return True

    def fade_out(self):
        """
        Decreases opacity until transparent
        """
        opacity = self.get_opacity()

        if opacity == 0:
            Gtk.Window.hide(self)
            self.image.set_from_pixbuf(None)
            self._fade_out_id = None

            return False

        self.set_opacity(opacity - 0.1)

        return True

    def cross_fade(self, pixbuf, next_pixbuf, duration):
        """
        Fades between two cover images

        :param pixbuf: the current cover image pixbuf
        :type pixbuf: :class:`GdkPixbuf.Pixbuf`
        :param next_pixbuf: the cover image pixbuf to fade to
        :type next_pixbuf: :class:`GdkPixbuf.Pixbuf`
        :param duration: the overall time for the fading
        :type duration: int
        """
        if self._cross_fade_step < duration:
            width, height = pixbuf.get_width(), pixbuf.get_height()
            alpha = (255 / duration) * self._cross_fade_step

            next_pixbuf.composite(
                dest=pixbuf,
                dest_x=0,
                dest_y=0,
                dest_width=width,
                dest_height=height,
                offset_x=0,
                offset_y=0,
                scale_x=1,
                scale_y=1,
                interp_type=GdkPixbuf.InterpType.BILINEAR,
                overall_alpha=int(alpha),
            )

            self.image.queue_draw()
            self._cross_fade_step += 1

            return True

        self._cross_fade_id = None
        self._cross_fade_step = 0

        return False

    def on_draw(self, widget, context):
        """
        Takes care of drawing the window
        transparently, if possible
        """

        context.set_source_rgba(1, 1, 1, 0)
        context.set_operator(cairo.OPERATOR_SOURCE)

        context.paint()

    def on_screen_changed(self, widget, event):
        """
        Updates the colormap
        """
        screen = widget.get_screen()
        visual = screen.get_rgba_visual() or screen.get_rgb_visual()
        self.set_visual(visual)

    def on_playback_track_start(self, type, player, track):
        """
        Updates the cover image and shows the window
        """
        self.set_cover_from_track(track)
        self.update_position()

    def on_playback_player_end(self, type, player, track):
        """
        Hides the window at the end of playback
        """
        self.hide()

    def on_cover_set(self, type, covers, track):
        """
        Updates the cover image after cover selection
        """
        self.set_cover_from_track(track)
        self.update_position()

    def on_cover_removed(self, type, covers, track):
        """
        Hides the window after cover removal
        """
        self.hide()

    def on_option_set(self, type, settings, option):
        """
        Updates the appearance
        """
        if option in (
            'plugin/desktopcover/anchor',
            'plugin/desktopcover/x',
            'plugin/desktopcover/y',
        ):
            self.update_position()
        elif option in (
            'plugin/desktopcover/override_size',
            'plugin/desktopcover/size',
        ):
            self.set_cover_from_track(player.PLAYER.current)


# vi: et sts=4 sw=4 tw=80



# Copyright (C) 2025  Johannes Sasongko <johannes sasongko org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


import os.path
import shelve
import tempfile

try:
    import berkeleydb
except ImportError:
    import bsddb3 as berkeleydb

from xl import common


def test_migration_to_sqlite():
    with tempfile.TemporaryDirectory(prefix="exaile-") as tmpdir:
        dbpath = os.path.join(tmpdir, "music.db")

        with shelve.BsdDbShelf(
            berkeleydb.hashopen(dbpath, 'c'), protocol=common.PICKLE_PROTOCOL
        ) as db:
            db['key'] = ({'tag': 'value'}, 2, {})

        with common.open_shelf(dbpath) as db:
            assert list(db.items()) == [('key', ({'tag': 'value'}, 2, {}))]



# Copyright (C) 2010 by Brian Parma
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import GObject


class CellRendererDays(Gtk.CellRendererText):
    '''Custom Cell Renderer for showing a ListView of 7 days with checkboxes, based off pygtk FAQ example'''

    __gtype_name__ = 'CellRendererDays'
    __gproperties__ = {
        'days': (object, 'days', 'List of enabled days', GObject.ParamFlags.READWRITE)
    }
    __gsignals__ = {
        'days-changed': (GObject.SignalFlags.RUN_FIRST, None, (str, object))
    }
    property_names = list(__gproperties__.keys())

    def __init__(self):
        Gtk.CellRendererText.__init__(self)
        self.model = Gtk.ListStore(bool, str)
        self.view = None
        self.view_window = None

        for day in [
            'Sunday',
            'Monday',
            'Tuesday',
            'Wednesday',
            'Thursday',
            'Friday',
            'Saturday',
        ]:
            self.model.append([True, day])

        self.set_property('text', 'Edit me')

    def _create_view(self, treeview):
        '''Create the Window and View to display when editing'''
        self.view_window = Gtk.Window()
        self.view_window.set_decorated(False)
        self.view_window.set_property('skip-taskbar-hint', True)

        self.view = Gtk.TreeView()

        self.view.set_model(self.model)
        self.view.set_headers_visible(False)

        cr = Gtk.CellRendererToggle()
        cr.connect('toggled', self._toggle)
        col = Gtk.TreeViewColumn('Enabled', cr, active=0)
        self.view.append_column(col)

        cr = Gtk.CellRendererText()
        col = Gtk.TreeViewColumn('Day', cr, text=1)
        self.view.append_column(col)

        # events
        self.view.connect('focus-out-event', self._close)
        self.view.connect('key-press-event', self._key_pressed)

        # should be automatic
        self.view_window.set_modal(False)
        self.view_window.set_transient_for(None)  # cancel the modality of dialog
        self.view_window.add(self.view)

        # necessary for getting the (width, height) of calendar_window
        self.view.show()
        self.view_window.realize()

    def do_set_property(self, pspec, value):
        '''Set property overload'''
        setattr(self, pspec.name, value)

    def do_get_property(self, pspec):
        '''Get property overload'''
        return getattr(self, pspec.name)

    def do_start_editing(
        self, event, treeview, path, background_area, cell_area, flags
    ):
        '''Called when user starts editing the cell'''

        if not self.get_property('editable'):
            return

        # create window/view if it doesn't exist
        if not self.view_window:
            self._create_view(treeview)
        else:
            self.view_window.show()

        # set display to reflect 'days' property
        for i, row in enumerate(self.model):
            row[0] = self.days[i]

        # position the popup below the edited cell (and try hard to keep the popup within the toplevel window)
        tree_x, tree_y = treeview.get_bin_window().get_origin()[1:]
        tree_w, tree_h = treeview.get_window().get_geometry()[2:4]
        my_w, my_h = self.view_window.get_window().get_geometry()[2:4]
        x = tree_x + min(cell_area.x, tree_w - my_w + treeview.get_visible_rect().x)
        y = tree_y + min(cell_area.y, tree_h - my_h + treeview.get_visible_rect().y)
        self.view_window.move(x, y)

        # save the path so we can return it in _done, and we aren't using dialog so we can't block....
        self._path = path

        return None  # don't return any editable, our Gtk.Dialog did the work already

    def _done(self):
        '''Called when we are done editing'''
        days = [row[0] for row in self.model]

        if days != self.days:
            self.emit('days-changed', self._path, days)

        self.view_window.hide()

    def _key_pressed(self, view, event):
        '''Key pressed event handler, finish editing on Return'''
        # event == None for day selected via double-click
        if (
            not event
            or event.type == Gdk.EventType.KEY_PRESS
            and Gdk.keyval_name(event.keyval) == 'Return'
        ):
            self._done()
            return True

    def _toggle(self, cell, path):
        '''Checkbox toggle event handler'''
        active = self.model[path][0]
        self.model[path][0] = not active
        return True

    def _close(self, view, event):
        '''Focus-out-event handler'''
        self._done()
        return True



# Copyright (C) 2012 Mathias Brodala
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os
import sys

from xl.nls import gettext as _
from xlgui.preferences import widgets

name = _('On Screen Display')
basedir = os.path.dirname(os.path.realpath(__file__))
ui = os.path.join(basedir, "osd_preferences.ui")
icon = 'dialog-information'


OSDPLUGIN = None


def page_enter(_preferencesdialog):
    """
    Shows a preview of the OSD
    """
    OSDPLUGIN.make_osd_editable(True)


def page_leave(_preferencesdialog):
    """
    Hides the OSD preview
    """
    OSDPLUGIN.make_osd_editable(False)


class ShowProgressPreference(widgets.CheckPreference):
    name = 'plugin/osd/show_progress'
    default = True


class DisplayDurationPreference(widgets.SpinPreference):
    name = 'plugin/osd/display_duration'
    default = 4


class BackgroundPreference(widgets.RGBAButtonPreference):
    name = 'plugin/osd/background'
    default = 'rgba(18, 18, 18, 0.8)'

    def __init__(self, preferences, widget):
        widgets.RGBAButtonPreference.__init__(self, preferences, widget)
        if sys.platform.startswith("win32"):
            # Setting opacity on Windows crashes with segfault,
            # see https://bugzilla.gnome.org/show_bug.cgi?id=674449
            widget.set_use_alpha(False)


class FormatPreference(widgets.TextViewPreference):
    name = 'plugin/osd/format'
    default = _(
        '<span font_desc="Sans 11" foreground="#fff">$title</span>\n'
        'by $artist\n'
        'from $album'
    )


class BorderRadiusPreference(widgets.SpinPreference):
    name = 'plugin/osd/border_radius'
    default = 10

    def _get_value(self):
        return self.widget.get_value_as_int()



# Copyright (C) 2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from xl import xdg, main, metadata, settings
from xl.nls import gettext as _
from xlgui.preferences import widgets
from xlgui import progress
from xl.common import SimpleProgressThread

name = _('Collection')
icon = 'folder-music'
ui = xdg.get_data_path('ui', 'preferences', 'collection.ui')


def _get_default_strip_list():
    return []
    # FIXME:  currently, this is broken by the backend not also having access
    # to the default set here, so we will just NOT set one.

    # TRANSLATORS: Grammatical articles that are ignored while sorting the
    # collection panel. For example, in French locales this could be
    # the space-separated list "l' la le les".
    # If this practice is not common in your locale, simply
    # translate this to string with single space.
    default_strip_list = _("the")
    return [v.lower() for v in default_strip_list.split(' ') if v]


class CollectionStripArtistPreference(widgets.ListPreference):
    default = _get_default_strip_list()
    name = 'collection/strip_list'

    def __init__(self, preferences, widget):
        widgets.ListPreference.__init__(self, preferences, widget)
        self.widget.connect('populate-popup', self._populate_popup_cb)

    def _get_value(self):
        """
        Get the value, overrides the base class function
        because we don't need shlex parsing. We actually
        want values like "l'" here.
        """
        values = [v.lower() for v in self.widget.get_text().split(' ') if v]
        return values

    def _populate_popup_cb(self, entry, menu):
        from gi.repository import Gtk

        entry = Gtk.MenuItem.new_with_mnemonic(_('Reset to _Defaults'))
        entry.connect('activate', self._reset_to_defaults_cb)
        entry.show()

        sep = Gtk.SeparatorMenuItem()
        sep.show()

        menu.attach(entry, 0, 1, 0, 1)
        menu.attach(sep, 0, 1, 1, 2)

    def _reset_to_defaults_cb(self, item):
        self.widget.set_text(' '.join(_get_default_strip_list()))


class FileBasedCompilationsPreference(widgets.CheckPreference):
    default = True
    name = 'collection/file_based_compilations'


class WriteRatingToAudioFileMetadataPreference(widgets.CheckPreference):
    default = False
    name = 'collection/write_rating_to_audio_file_metadata'


class WriteRatingToAudioFileMetadataPOPMemail(widgets.Preference, widgets.Conditional):
    default = ""
    name = 'collection/write_rating_to_audio_file_metadata_popm_mail'
    condition_preference_name = 'collection/write_rating_to_audio_file_metadata'

    def __init__(self, preferences, widget):
        widgets.Preference.__init__(self, preferences, widget)
        widgets.Conditional.__init__(self)

    def on_check_condition(self):
        return self.condition_widget.get_active() is True


class WriteRatingToAudioFileMetadataSyncNow(widgets.Button, widgets.Conditional):
    default = ""
    name = "collection/write_rating_to_audio_file_metadata_sync_now"
    condition_preference_name = 'collection/write_rating_to_audio_file_metadata'

    def __init__(self, preferences, widget):
        widgets.Button.__init__(self, preferences, widget)
        widgets.Conditional.__init__(self)
        self.progress = None
        self.scan_thread = None
        self.monitor = None

    def on_check_condition(self):
        return self.condition_widget.get_active() is True

    def on_clicked(self, button):
        if self.progress:
            return

        curr_page = self.preferences.last_page
        box = self.preferences.builders[curr_page].get_object(
            'collection/write_rating_to_audio_file_metadata_progress'
        )
        self.progress = progress.ProgressManager(box)

        self.scan_thread = SimpleProgressThread(
            self.track_scan,
        )
        self.scan_thread.connect('done', self.on_done)
        self.monitor = self.progress.add_monitor(
            self.scan_thread,
            _("Migrating ratings to audio file metadata"),
            'document-open',
        )

        box.show()

    def track_scan(self):
        exaile = main.exaile()
        collection = exaile.collection
        total = len(collection.tracks)
        i = 0

        for track in collection.tracks:
            trax = collection.tracks[track]
            format_data = metadata.get_format(track)

            if '__rating' in format_data.tag_mapping:
                rating = trax.get_rating()
                if rating > 0:
                    trax.set_rating(rating)

            i += 1
            yield i, total

    def on_done(self, a):
        self.progress.remove_monitor(self.monitor)
        self.progress = None


class UseLegacyMetadataMappingPreference(widgets.CheckPreference):
    default = False
    name = 'collection/use_legacy_metadata_mapping'


class UseLegacyMetadataMappingSyncNow(widgets.Button, widgets.Conditional):
    default = ""
    name = "collection/use_legacy_metadata_mapping_sync_now"
    condition_preference_name = 'collection/use_legacy_metadata_mapping'

    def __init__(self, preferences, widget):
        widgets.Button.__init__(self, preferences, widget)
        widgets.Conditional.__init__(self)
        self.progress = None
        self.scan_thread = None
        self.monitor = None

    def on_check_condition(self):
        return self.condition_widget.get_active() is True

    def on_clicked(self, button):
        if self.progress:
            return

        curr_page = self.preferences.last_page
        box = self.preferences.builders[curr_page].get_object(
            'collection/use_legacy_metadata_mapping_progress'
        )
        self.progress = progress.ProgressManager(box)

        self.scan_thread = SimpleProgressThread(
            self.track_scan,
        )
        self.scan_thread.connect('done', self.on_done)
        self.monitor = self.progress.add_monitor(
            self.scan_thread,
            _("Migrating metadata"),
            'document-open',
        )

        box.show()

    def track_scan(self):
        from xl.metadata import flac, ogg

        exaile = main.exaile()
        collection = exaile.collection
        total = len(collection.tracks)
        i = 0

        for track in collection.tracks:
            trax = collection.tracks[track]
            format_data = metadata.get_format(track)

            if not isinstance(format_data, flac.FlacFormat) and not isinstance(
                format_data, ogg.OggFormat
            ):
                continue

            settings.set_option('collection/use_legacy_metadata_mapping', True)
            bpm = trax.get_tag_disk('tempo')
            comment = trax.get_tag_disk('description')

            settings.set_option('collection/use_legacy_metadata_mapping', False)
            trax.set_tag_disk('bpm', bpm)
            trax.set_tag_disk('comment', comment)

            i += 1
            yield i, total

    def on_done(self, a):
        self.progress.remove_monitor(self.monitor)
        self.progress = None
        settings.set_option('collection/use_legacy_metadata_mapping', False)
        self.condition_widget.set_active(False)


# vim:ts=4 et sw=4



# Copyright (C) 2010 Johannes Schwarz
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 1, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import dbus
from gi.repository import GLib
from gi.repository import Gtk

from xl import event, providers, settings
from xl.nls import gettext as _
from xlgui.widgets import dialogs, menu

from . import shutdown_preferences

SHUTDOWN = None


class Shutdown:
    def __init__(self, exaile):
        self.exaile = exaile
        self.do_shutdown = False
        self.do_close = settings.get_option(
            "shutdown/activate_closing_by_default", False
        )

        # add menuitem to tools menu
        providers.register(
            'menubar-tools-menu',
            menu.simple_separator('plugin-sep-shutdown', ['slow-scan-collection']),
        )

        item = menu.check_menu_item(
            'close',
            ['plugin-sep-shutdown'],
            _('Close Exaile after Playback'),
            #   checked func                # callback func
            lambda *x: self.do_close,
            lambda w, n, p, c: self.on_toggle(w, n),
        )
        providers.register('menubar-tools-menu', item)

        if self.do_close:
            event.add_ui_callback(self.on_playback_player_end, 'playback_player_end')

        item = menu.check_menu_item(
            'shutdown',
            ['close'],
            _('Shutdown after Playback'),
            #   checked func                # callback func
            lambda *x: self.do_shutdown,
            lambda w, n, p, c: self.on_toggle(w, n),
        )
        providers.register('menubar-tools-menu', item)

        self.countdown = None
        self.counter = 10

        self.message = dialogs.MessageBar(
            parent=exaile.gui.builder.get_object('player_box'),
            buttons=Gtk.ButtonsType.CLOSE,
        )
        self.message.connect('response', self.on_response)

    def on_toggle(self, menuitem, name):
        if menuitem.get_active() and name == 'close':
            self.do_close = True
            self.do_shutdown = False
            self.message.show_info(
                _('Close scheduled'),
                _('Exaile will be closed at the end of playback.'),
            )
        elif menuitem.get_active() and name == 'shutdown':
            self.do_close = False
            self.do_shutdown = True
            self.message.show_info(
                _('Shutdown scheduled'),
                _('Computer will be shutdown at the end of playback.'),
            )
        else:
            self.disable_all()

        if self.do_close or self.do_shutdown:
            event.add_ui_callback(self.on_playback_player_end, 'playback_player_end')

    def disable_all(self):
        self.do_shutdown = False
        self.do_close = False
        event.remove_callback(self.on_playback_player_end, 'playback_player_end')

        # Stop possible countdown
        if self.countdown is not None:
            GLib.source_remove(self.countdown)
            self.countdown = None

        # Prepare for a new run
        self.counter = 10

        # Reset message button layout
        self.message.hide()
        self.message.clear_buttons()
        self.message.add_button(Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)

    def on_playback_player_end(self, event, player, track):
        """
        Tries to shutdown the computer
        """
        self.message.set_message_type(Gtk.MessageType.INFO)
        if self.do_close:
            self.message.set_markup(_('Imminent Closing'))
        else:
            self.message.set_markup(_('Imminent Shutdown'))
        self.message.clear_buttons()
        self.message.add_button(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL)

        if self.countdown is not None:
            GLib.source_remove(self.countdown)

        self.counter = settings.get_option("shutdown/timeout", 10)
        self.countdown = GLib.timeout_add_seconds(1, self.on_timeout)

    def on_response(self, widget, response):
        """
        Cancels shutdown if requested
        """
        if response == Gtk.ResponseType.CANCEL:
            self.disable_all()

    def on_timeout(self):
        """
        Tries to shutdown the computer
        """
        if self.counter > 0:
            msg_close = _('Exaile will be closed in %d seconds.') % self.counter
            msg_shutdown = (
                _('The computer will be shut down in %d seconds.') % self.counter
            )

            if self.do_close:
                msg = msg_close
            elif self.do_shutdown:
                msg = msg_shutdown
            else:
                return False

            self.message.set_secondary_text(msg)
            self.message.show()

            self.counter -= 1

            return True

        if self.do_close:
            self.do_close = False
            self.exaile.quit()
            return  # Should not be necessary

        self.do_shutdown = False

        bus = dbus.SystemBus()

        try:
            proxy = bus.get_object('org.freedesktop.login1', '/org/freedesktop/login1')
            proxy.PowerOff(False, dbus_interface='org.freedesktop.login1.Manager')
        except dbus.exceptions.DBusException:
            try:
                proxy = bus.get_object(
                    'org.freedesktop.ConsoleKit', '/org/freedesktop/ConsoleKit/Manager'
                )
                proxy.Stop(dbus_interface='org.freedesktop.ConsoleKit.Manager')
            except dbus.exceptions.DBusException:
                try:
                    proxy = bus.get_object(
                        'org.freedesktop.Hal', '/org/freedesktop/Hal/devices/computer'
                    )
                    proxy.Shutdown(
                        dbus_interface='org.freedesktop.Hal.Device.SystemPowerManagement'
                    )
                except dbus.exceptions.DBusException:
                    self.message.show_warning(
                        _('Shutdown failed'),
                        _('Computer could not be shutdown using D-Bus.'),
                    )

    def destroy(self):
        """
        Cleans up
        """
        if self.countdown is not None:
            GLib.source_remove(self.countdown)

        event.remove_callback(self.on_playback_player_end, 'playback_player_end')
        for item in providers.get('menubar-tools-menu'):
            if item.name == 'shutdown' or item.name == 'close':
                providers.unregister('menubar-tools-menu', item)


def enable(exaile):
    if exaile.loading:
        event.add_callback(_enable, 'exaile_loaded')
    else:
        _enable(None, exaile, None)


def _enable(eventname, exaile, nothing):
    global SHUTDOWN
    SHUTDOWN = Shutdown(exaile)


def disable(exaile):
    global SHUTDOWN
    SHUTDOWN.destroy()


def get_preferences_pane():
    return shutdown_preferences
